import logging
from .utility import *

LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
DEFAULT_LEVEL = "info"

FORMAT = (
    "%(asctime)s [%(filename)s|%(funcName)s] [%(levelname)s] [%(lineno)s] %(message)s"
)

LOG_FILE_DEFAULT = "pipet_default.log"

def parse_log_level(log_level):
    return LEVELS.get(log_level, LEVELS[DEFAULT_LEVEL])

def setup_logging(log_level, log_path='./'):
    logger = logging.getLogger('__default__')
    logger.setLevel(parse_log_level(log_level))
    
    formatter = logging.Formatter(FORMAT)
    logger.addHandler(get_console_handler(formatter))
    logger.addHandler(get_file_handler(formatter, 
                                       log_path+LOG_FILE_DEFAULT))
    #  propagate the error up to parent
    logger.propagate = False




from .base_object import BaseObject
from .customer import Customer
from pypipet.core.sql.query_interface import get_product_by_destination
import re

class OrderItem(BaseObject):
    __table_name__ = 'order_item'
    def __init__(self, attrs: dict = None):
        super().__init__(attrs)
    
    def set_order_item(self, table_objs, attrs: dict):
        self.set_attrs_by_table_class(table_objs.get(self.__table_name__), 
                                     attrs)
    def get_order_item_info(self, **kwargs):
        table_objs = kwargs.get('table_objs')
        session = kwargs.get('session')
        params = {}
        if self.destination_id is not None:
            params['id'] = self.destination_id
        else:
            params.update({
                        'front_shop_id':  self.front_shop_id,
                        'destination_product_id':  str(self.destination_product_id)
                    })
        product_info = get_product_by_destination(
                table_objs, 
                session, 
                params
                )
        product_info.update(params)
        return product_info



class Order(BaseObject):
    __table_name__ = 'shop_order'
    def __init__(self, attrs: dict = None):
        super().__init__(attrs)
        self.order_items = []
        self.billing_customer = None
        self.shipping_customer = None
        self.trackings = None

    def set_order(self, table_objs, attrs: dict):
        self.set_attrs_by_table_class(table_objs.get(self.__table_name__), 
                                     attrs)
        if attrs.get('order_item'):
            for item_data in attrs['order_item']:
                item = OrderItem()
                item.set_attrs_by_table_class(
                            table_objs.get(item.__table_name__), 
                            item_data)
                self.order_items.append(item)
        if attrs.get('shipping_customer'):
            customer = Customer()
            customer.set_attrs_by_table_class(
                            table_objs.get(customer.__table_name__), 
                            attrs['shipping_customer'])
            self.shipping_customer = customer

        if attrs.get('billing_customer'):
            customer = Customer()
            customer.set_attrs_by_table_class(
                            table_objs.get(customer.__table_name__), 
                            attrs['billing_customer'])
            self.billing_customer = customer

        if attrs.get('fulfillment'):
            self.trackings = attrs['fulfillment']

    def update_order_id(self, shop_order_id):      
        self.set_attr('id', shop_order_id) 
        for item in self.order_items:
            item.set_attr('shop_order_id', shop_order_id)
        if self.trackings:
            for i, tracking in enumerate(self.trackings):
                self.trackings[i]['shop_order_id'] = shop_order_id

    def get_order_items(self):
        pass

    def get_shipping(self):
        pass

    def get_customer(self):
        pass


    def get_issues(self):
        pass

    def get_order_info(self, **kwargs):
        info = self.get_all_attrs()
        info['order_items'] = []
        for item in self.order_items:
            if kwargs.get('item_info') == True:
                info['order_items'].append(item.get_order_item_info(**kwargs))
            else:
                info['order_items'].append(item.get_all_attrs())
        info['billing'] = self.billing_customer.get_all_attrs()
        info['shipping'] = self.shipping_customer.get_all_attrs()
        return info

# class Fulfillment(BaseObject):
#     __table_name__ = 'fulfillment'
#     def __init__(self, attrs: dict = None):
#         super().__init__(attrs)

#     def set_fulfillment(self, table_objs, attrs: dict):
#         self.set_attrs_by_table_class(table_objs.get(self.__table_name__), 
#                                      attrs)

#     def update_status(self):
#         pass

#     def update_tracking(self, tracking: dict=None):
#         if self.get_attr('tracking_id') is None:
#             self.tracking_id = tracking.get('tracking_id', None)
#             self.provider = tracking.get('provider', None)
#         else:
#             if tracking.get('tracking_id', None) != self.tracking_id:
#                 self.tracking_id = tracking.get('tracking_id', None)
#                 self.provider = tracking.get('provider', None)

#         # return self.validate_tracking_id()

#     def validate_tracking_id(self):
#         return
                    


        


import click 
from .cli import cli
# from pypipet.api.app import create_app

@cli.command()
@click.argument("action",  type=click.Choice(['start', 'stop']))
@click.pass_context
def api(ctx, action):
    """
    run flask api
    """
    click.echo('running flask api')
    project = ctx.obj['project']
    config = ctx.obj["config"]
    if config is None:
        click.secho('create project first',
                   fg='yellow')
        return

    project.initialize_project(config)
    # app = create_app(project)
    # app.run(debug=True)

    


import os
from .db import project_engine, import_data
from .config_service import *
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.automap import automap_base
from pypipet.plugins.woocommerce.api import API as WC_API
import shopify as SHOPIFY_API
from pypipet.core.logging import setup_logging
from pypipet.core.fileIO.file_loader import read_yml_file
from pypipet.core.fileIO.file_saver import write_yaml, copy_file
from pypipet.core.shop_conn.shop_connector_wc import WCShopConnector
from pypipet.core.shop_conn.shop_connector_shopify import SPFShopConnector
from pypipet.core.shop_conn.spf_auth import start_session
from .config import *
import logging
logger = logging.getLogger('__default__')

class PipetContext():
    def __init__(self, *args, **kwargs):
        self.engine = kwargs.get('engine', None)
        self.table_objects = kwargs.get('engine', None)
        self.root = kwargs.get('root_path', os.getcwd())
        if not self.root.endswith('/'):
            self.root += '/'
        self.config = None
        self.db_config = None
        self._shop_connectors = {}
        self.schema = ''

    def start_project(self, project_name, **kwargs):
        if os.path.isfile(self.root + 'setting.yaml'):
            return False 
        project_setting = read_yml_file(PROJECT_SETTING)
        project_setting['project_setting']['name'] = project_name
        project_setting['project_setting']['home_dir'] = self.root
        # project_setting['project_setting']['attr_list'] = ['brand', 'upc', 'size', 'color']

        write_yaml(project_setting,
                   self.root + 'setting.yaml')

        os.makedirs(self.root + 'bundle', exist_ok=True)
        copy_file(DB_SETTING, self.root + 'bundle/db_setting.yaml')

        copy_file(WC_MAPPING,
                  self.root + 'bundle/wc_field_mapping.yaml')

        copy_file(SPF_MAPPING,
                   self.root + 'bundle/spf_field_mapping.yaml')

        copy_file(FILE_TEMPLATE,
                   self.root + 'bundle/file_template.yaml')
        return True

    def initialize_project(self, config_file=None, validate_db=False):
        if self.config is None:
            self.config = read_yml_file(config_file)['project_setting']
            
        self.set_log_level(self.config['log_level'], 
                           path=self.config['log_path'])

        if self.db_config is None:
            self.db_config = read_yml_file(
                     self.config['home_dir'] \
                    + self.config['databse_setting'])
            if validate_db:
                is_valid, message = validate(self.db_config)

                if not is_valid:
                    logger.info(message)
                    return 
        
        if self.db_config['db_setting']['db_type'] in  ('postgres'):
            self.schema = self.db_config['db_setting']['db_conn'].get('schema', '')   
        
        if self.engine is None:
            self.engine = project_engine(self.db_config)

        if self.engine:
            self.set_table_objects(self.engine)
            self._session_maker = sessionmaker(bind=self.engine)
            self._initialize_shops()
        

    def _initialize_shops(self):
        for shop_name, shop_config in self.config['shops'].items():
            if shop_config['site_type'] == 'wc':
                shop_conn = WCShopConnector(shop_name, 
                            shop_config['site_type'],
                            batch_size=shop_config['batch_size']) 
                self.set_wc_mapping()
                api = WC_API(
                    url=shop_config['url'],
                    consumer_key=shop_config['consumer_key'],
                    consumer_secret=shop_config['consumer_secret'],
                    version=shop_config['version']
                  )

                shop_conn.set_shop_api(api)
                self.set_shop_connector(shop_name, shop_conn)
            elif shop_config['site_type'] == 'shopify':
                is_private_app = shop_config.get('private_app_password') is not None
                api_session = start_session(
                                            private_app=is_private_app,
                                            **shop_config)
                
                shop_conn = SPFShopConnector(shop_name,
                            shop_config['site_type'])
                shop_conn.set_shop_api(api_session)

                self.set_spf_mapping()
                self.set_shop_connector(shop_name, shop_conn)
                # shopify.ShopifyResource.activate_session(session)

    def get_session_maker(self):
        if self.engine:
            if self._session_maker:
                return self._session_maker
            self._session_maker = sessionmaker(bind=self.engine)
            return self._session_maker
        return None
    
    def set_log_level(self, log_level, path='./'):
        if log_level is None: log_level = 'info'
        if path is None: path = './'
        self.log_level = log_level
        setup_logging(log_level, log_path=path)


    def set_table_objects(self, engine):
        base = automap_base()
        if self.schema == '':
            base.prepare(engine, reflect=True)
        else:
            base.prepare(engine, reflect=True, schema=self.schema)
        self.table_objects = base.classes 
       

    def get_table_objects(self):
        return self.table_objects

    def get_table_columns(self, table_name):
        if self.table_objects.get(table_name):
            return self.table_objects.get(table_name).__table__.columns.keys()

    
    def set_shop_connector(self, name, connector):
        if connector.shop_type == 'wc':
            connector.set_field_mapping(self.get_wc_mapping())
        elif connector.shop_type == 'shopify':
            connector.set_field_mapping(self.get_spf_mapping())
        self._shop_connectors[name] = connector

    def get_shop_connector(self, name):
        return self._shop_connectors.get(name, None)

    def set_wc_mapping(self):
        self._wc_mapping = load_mapping(
                 filename=self.root + self.config['wc_field_mapping'])

    def get_wc_mapping(self):
        return self._wc_mapping.get('mapping')

    def set_spf_mapping(self):
        self._spf_mapping = load_mapping(
                 filename=self.root + self.config['spf_field_mapping'])

    def get_spf_mapping(self):
        return self._spf_mapping.get('mapping')

    def import_static_data(self):
        session = self._session_maker()
        settings = self.db_config
                
        if settings and settings.get('manual_data_input'):
            import_data(self.get_table_objects(), 
                       session, 
                       settings.get('manual_data_input'))
        session.close()
    

    def load_file_template(self):
        return read_yml_file(self.config['file_template'])
    
    


import click 
from .cli import cli
from pypipet.core.operations.utility import get_front_shop_id
from pypipet.core.operations.frontshop import add_taxonomy_to_db
from pypipet.core.operations.frontshop import add_product_to_db_bulk
from pypipet.core.operations.frontshop import add_product_to_db
from pypipet.core.operations.frontshop import add_destination_to_db
from pypipet.core.operations.frontshop import load_products_from_shop
from pypipet.core.fileIO.file_loader import read_csv_file
from .utility import col2dict
from pprint import pprint

@cli.command()
@click.argument("source_type", type=click.Choice(["shop", "file"]))
@click.option("--shop", 
                help="frontshop name (in setting.shops)")
@click.option("-f", "--filename", default='',
                help="product feed csv")
@click.option("--currency", default='USD') 
@click.pass_context
def catalog(ctx,source_type, shop, filename, currency):
    """
    import catalog from frontshop or files
    for file import, (generate template first)
    """
    click.echo('loading products from {}...'.format(source_type))
    project = ctx.obj['project']
    config = ctx.obj["config"]
    if config is None:
        click.secho('create project first',
                   fg='yellow')
        return

    project.initialize_project(config)
    table_classes = project.get_table_objects()
    sessionmaker = project.get_session_maker()
    session = sessionmaker()

    shop_conn = project.get_shop_connector(shop)
    if shop_conn is None:
        click.secho('shop setting not found: {}'.format(shop),
                   fg='yellow')
        session.close()
        return 
    
    get_front_shop_id(table_classes, session, shop_conn)
   
    if source_type == 'shop':
        #add taxonomy
        # click.echo('sync categories...')
        # add_taxonomy_to_db(table_classes.get('category'), 
        #                    session, 
        #                    shop_conn)
        #pull products
        _load_from_shop_batch(table_classes, session, 
                         shop_conn, 
                         default_brand=project.config['default_brand'],
                         currency=currency)

    elif source_type == 'file':
        _load_from_file(table_classes, session, filename,
                         shop_conn, 
                         default_brand=project.config['default_brand'],
                         currency=currency) 
    else:
        click.secho('invalid source to load products',
                   fg='yellow')
    session.close()

def _load_from_shop_batch(table_objs, session, 
                         shop_connector, default_brand='',
                         currency='USD'):
    load_products_from_shop(table_objs, session, shop_connector, 
                         start_from=1, currency='USD')
        

def _load_from_file(table_objs, session, filename,
                         shop_connector, default_brand='',
                         currency='USD'):
    data = read_csv_file(filename)
    must_exist_fields = ['product.category', 
                         'product.product_name',
                         'variation.sku']
    
    #validation
    for i, row in data.iterrows():
        sku = str(row['variation.sku'])
        # click.echo('row {}: sku {}'.format(i, sku))
        for key in must_exist_fields:
            if row[key].strip() == '':
                click.secho('missing {} at {}, stop importing'\
                                 .format(key, sku), 
                        fg='yellow')
                return 
    
    products = []
    for i, row in data.iterrows():
        sku = str(row['variation.sku'])
        click.echo('row {}: sku {}'.format(i, sku))
        p = col2dict(sku, row, default_brand)
        products.append(p)
        
    click.echo('editing database')
    add_product_to_db_bulk(table_objs, session, shop_connector, 
                          products, currency=currency)


        


from dataclasses import fields
from email import message
from turtle import up
import shopify
from shopify.resources import fulfillment_service
from pyactiveresource.activeresource import ActiveResource
from shopify.resources.custom_collection import CustomCollection
import logging
logger = logging.getLogger('__default__')

_ignore_status = ('cancelled', 'error', 'failure')

def add_request_handler(func):
    def wrapper(*args, **kwargs):
        try:
            res, suc = func(*args, **kwargs)
            if suc:
                return res.attributes, suc
            
            return res.errors.errors, suc
        except Exception  as e:
            logger.debug('requst error')
            logger.debug(e)
            return None, None
    return wrapper

def request_handler(func):
    def wrapper(*args, **kwargs):
        try:
            res = func(*args, **kwargs)
            if res is None:
                return 
            if isinstance(res, list):
                return [r.attributes for r in res]
            elif isinstance(res,ActiveResource):
                return res.attributes
            return res
            
        except Exception  as e:
            logger.debug('requst error')
            logger.debug(e)
            return 
    return wrapper

def update_request_handler(func):
    def wrapper(*args, **kwargs):
        try:
            res = func(*args, **kwargs)
            return res
            
        except Exception  as e:
            logger.debug('requst error')
            logger.debug(e)
            return 
    return wrapper


@add_request_handler
def _add_product(data: dict, status='active', publish_scope='web', **kwargs):
    data.update({
        'status': status,
        'publish_scope': publish_scope
    })
    if kwargs.get('template'): 
        data['template_suffix'] = kwargs['template']

    p = shopify.Product(attributes=data)
    suc = p.save()

    return p, suc

def add_product(data: dict, status='active', publish_scope='web', **kwargs):
    p, suc = _add_product(data, status, publish_scope, **kwargs)
    #unfold variants
    if p and p.get('variants'):
        for i, vari in enumerate(p['variants']):
            if isinstance(vari, ActiveResource):
                if kwargs.get('variant_imgs') and \
                kwargs['variant_imgs'].get(vari.sku) is not None:
                    vari.image_id = p['images'][kwargs['variant_imgs'][vari.sku]].id
                    vari.save()
                p['variants'][i] = vari.attributes
    return p, suc

@request_handler
def add_product_to_collection(product_id, collection_id):
    res = shopify.Collect.create({"collection_id": collection_id, 
                         "product_id": product_id})
    return res
 
def add_products_bulk():
    pass

@request_handler
def find_product(product_id=None):
    if product_id is not None:
        return shopify.Product.find(product_id)
    else:
        return shopify.Product.find()

def get_product_info(product_id, **kwargs):
    info = find_product(product_id)
    if info:
        if  info.get('variants'):
            for i, vari in enumerate(info['variants']):
                info['variants'][i] = vari.attributes
                if kwargs.get('include_meta') == True:
                    info['variants'][i]['meta'] = find_metafield(
                    mId=None, resource={
                        'resource_id': vari.id,
                        'resource': f'products/{product_id}/variants'
                    })
        if  info.get('images'):
            for i, img in enumerate(info['images']):
                info['images'][i] = img.attributes
        if  info.get('options'):
            for i, option in enumerate(info['options']):
                info['options'][i] = option.attributes
        if  info.get('image'):
            info['image'] = info['image'].attributes
        if kwargs.get('include_meta') == True:
            # resource ={
            #     'resource_id': 42331660189866,
            #     'resource': 'products/7146500587690/variants'
            # }
            info['meta'] = find_metafield(mId=None, resource={
                    'resource_id': product_id,
                    'resource': 'products'
            })
        return info
    
@request_handler   
def get_products(**kwargs):
    return shopify.Product.find(**kwargs)

def get_all_products(**kwargs):
    products = get_all_items(shopify.Product,  **kwargs)
    return sorted(products, key=lambda d: d['id'])




def get_next_page(obj: ActiveResource=None, first_page=False, 
                      next_url=None, **kwargs):
    if next_url:
        return obj.find(from_=next_url, **kwargs)
    if obj and first_page:
        return obj.find(**kwargs)
    if obj and obj.has_next_page():
        return obj.next_page(**kwargs)

@update_request_handler
def update_product(product_id, variant_id, data):
    p = shopify.Product.find(product_id) 
    product_updates = {}
    from pprint import pprint 
    pprint(data)
    for k, value in data.items():
        if k == 'images' and isinstance(value, list):
            product_updates[k] = p.images + value 
        elif k == 'variants':
            for variant_update in value:
                update_variant(variant_id, variant_update)
        elif k == 'options':
            product_updates[k] = p.options + value 
        elif k == 'destination':
            update_variant(variant_id, value)
        elif k in ('price', 'sku', 'qty', 'in_stock', 
                 'discount', 'compare_at_price', 'available'):
            pass
        elif not (isinstance(value, list) or isinstance(value, dict)):
            product_updates[k] = value
    if len(product_updates) > 0:
        p._update(product_updates)
        p.save()
    return p

def update_products_bulk():
    pass 

@add_request_handler
def add_variation(product_id, data: dict):
    #add image to product
    variant_img_id = None
    if data.get('images'):
        for img in data['images']:
            # print(img)
            new_img, suc = add_image(product_id, img)
            # print(new_img)
            if suc and variant_img_id is None:
                variant_img_id = new_img['id']

    data.update({'product_id': product_id,
                 'image_id': variant_img_id}) 
    variant = shopify.Variant(attributes=data)
    suc = variant.save()
    return variant, suc

@update_request_handler
def update_variant(variant_id, data, **kwargs):
    variant = shopify.Variant.find(variant_id, **kwargs)
    if data.get('available') is not None:
        if data['available'] == False:
            #set qty to 0
            update_inventory_level(variant.inventory_item_id, 
                                            {'in_stock': 0})
            return 
        #update inventory
        update_inventory_level(variant.inventory_item_id, 
                                        {'in_stock': data['in_stock']})
    variant._update(data)
    variant.save()
    return variant

@request_handler
def find_variation(variant_id=None, **kwargs):
    if variant_id is not None:
        return shopify.Variant.find(variant_id, **kwargs)
    else:
        return shopify.Variant.find(**kwargs)

@request_handler
def find_inventory_item(inv_item_id):
    return shopify.InventoryItem.find(inv_item_id)

@request_handler
def find_inventory_level(inv_item_id):
    return shopify.InventoryLevel.find(
        inventory_item_ids=inv_item_id)

@update_request_handler
def update_inventory_level(inv_item_id, data, **kwargs):
    qty = data.get('in_stock')
    if qty is None: qty = data.get('qty')
    if qty is None: 
        logger.debug('invalid quantity')
        return 

    #update variant-inventory_management,fulfillment_service
    if kwargs.get('set_inventory_management') == True \
    and kwargs.get('variant_id'):
        variant = shopify.Variant.find(kwargs['variant_id'])
        inv_item_id = variant.inventory_item_id
        variant._update({
            'inventory_management': kwargs.get('inventory_management','shopify'),
            'fulfillment_service': kwargs.get('fulfillment_service','manual')
        })
        variant.save()

    if inv_item_id is None and kwargs.get('variant_id'):
        variant = shopify.Variant.find(kwargs['variant_id'])
        inv_item_id = variant.inventory_item_id

    items = shopify.InventoryLevel.find(
        inventory_item_ids=inv_item_id)
    
    if items:
        for item in items:
            item.set(item.location_id, inv_item_id, 
                qty, **kwargs)
            
        return items

def get_inventory(inv_item_id):
    inv_item =  find_inventory_item(inv_item_id)
    if inv_item:
        inv_level = find_inventory_level(inv_item_id)
        if inv_level:
            inv_item.update(inv_level)
        return inv_item

# @update_request_handler
# def update_inventory(inv_item_id, data,**kwargs):
#     inv_item = shopify.InventoryItem.find(inv_item_id)
#     inv_item._update(data)
#     inv_item.save()

#     update_inventory_level(inv_item_id, data, **kwargs)



@request_handler
def find_location(location_id=None):
    if location_id is not None:
        return shopify.Location.find(location_id)
    else:
        return shopify.Location.find()

def add_order():
    pass 

def add_orders_bulk():
    pass

def get_all_orders(**kwargs):
    print(kwargs)
    orders = get_all_items(shopify.Order,  **kwargs)
    orders = sorted(orders, key=lambda d: d['id'])
    return [extract_order_info(order) for order in orders ]

@request_handler
def find_order(order_id):
    return shopify.Order.find(order_id) 

def _extract_order_tax(lines):
    for i, item in enumerate(lines):
        lines[i] = item.attributes
        if lines[i].get('tax_lines'):
            lines[i]['tax_lines'] = [t.attributes \
                  for t in lines[i]['tax_lines']]


def extract_order_info(order_info):
    if order_info.get('billing_address') is not None:
        order_info['billing_address'] = order_info['billing_address'].attributes
    if order_info.get('client_details') is not None:
        order_info['client_details'] = order_info['client_details'].attributes
    if order_info.get('customer') is not None:
        order_info['customer'] = order_info['customer'].attributes
    if order_info.get('shipping_address') is not None:
        order_info['shipping_address'] = order_info['shipping_address'].attributes

    if order_info.get('line_items') is not None:
        _extract_order_tax(order_info['line_items'])
    if order_info.get('shipping_lines') is not None:
        _extract_order_tax(order_info['shipping_lines'])

    if order_info.get('fulfillments') is not None:
        for i, ff in enumerate(order_info['fulfillments']):
            order_info['fulfillments'][i] = ff.attributes

    if order_info.get('refunds') is not None:
        for i, refund in enumerate(order_info['refunds']):
            refund_info = refund.attributes
            refund_info['amount'] = 0
            if refund_info.get('transactions'):
                for trans in refund_info['transactions']:
                    refund_info['amount'] += float(trans.amount)
            order_info['refunds'][i] = refund_info

    return order_info

def get_order_info(order_id):
    order = find_order(order_id) 
    return extract_order_info(order)

@update_request_handler
def update_order(order_id, order_data):
    order = shopify.Order.find(order_id)
    order._update(update_order)
    suc = order.save()
    return order, suc

@update_request_handler
def update_line_items(order_id, items, **kwargs):
    """not supported in shopify (after order has been created)"""
    pass
    # order = shopify.Order.find(order_id, fields='line_items')
    # line_items = order.line_items
    # for item in line_items:
    #     if items.get(str(item.variant_id)) is not None:
    #         from pprint import pprint
    #         print('-------------line_item',item.quantity)
    #         # item._update({'quantity': 1})
    #         item.quantity = 1
    #         suc = item.save()
    #         logger.debug(f'update line item {item.variant_id} in order {order_id}: {suc}')

@update_request_handler
def update_order_status(order_id, status):
    order = shopify.Order.find(order_id)
    did_update = False
    if order:
        if status == 'close':
            order.close()
            did_update = True
        elif status == 'open':
            order.open()   
            did_update = True 
        elif status == 'cancel':
            order.cancel()  
            did_update = True
    return did_update

@add_request_handler
def update_customer(order_id, customer_data, customer_id=None, **kwargs):
    """"shopify can not update billing address with api"""
    order = shopify.Order.find(order_id, fields='id,customer')
    if customer_id is None:
        customer_id = order.customer.id
    
    update_order = {}
    customer_update = {}
    if kwargs.get('is_billing'):
        # update_order['billing_address'] = customer_data
        if customer_data.get('email'): 
            update_order['email'] = customer_data['email']
            customer_update['email'] = customer_data['email']
        if customer_data.get('phone'): 
            update_order['phone'] = customer_data['phone']
            customer_update['phone'] = customer_data['phone']
    if kwargs.get('is_shipping'):
        update_order['shipping_address'] = customer_data
   
    if len(update_order) > 0:
        order._update(update_order)
        suc = order.save()
    
    if len(customer_update) > 0:
        customer = shopify.Customer.find(customer_id)
        if customer_data.get('address1') and customer_data.get('zip') \
        and customer_data.get('province') and customer_data.get('country'):
            
            customer_update['addresses'] = customer.addresses + [customer_data]
            
        customer._update(customer_update)
        suc = customer.save()

    return order, suc
    

@request_handler
def find_fulfillment(include_cancelled=False, **kwargs):
    ffs = shopify.Fulfillment.find(**kwargs)
    if ffs and not include_cancelled:
        filtered_ffs = []
        for ff in ffs:
            if ff.status != 'cancelled':
                filtered_ffs.append(ff)
        return filtered_ffs 

    return ffs



def _order_has_ff(ffs):
    if ffs is None or len(ffs) == 0:
        return False 
    for ff in ffs:
        if ff.status in _ignore_status:
            return False 

    return True
    

@update_request_handler
def update_fulfillment_tracking(order_id, ff_data, add_new=True, message=True, 
                                      ff_id=None, **kwargs):
    exists = shopify.Fulfillment.find(order_id=order_id)
    if add_new and not _order_has_ff(exists):
        ff = shopify.Fulfillment(
            attributes=ff_data, prefix_options={'order_id': order_id}
        )
        ff.save()
        return ff, True
    elif not add_new and exists and len(exists) > 0:
        did_update = False
        for ff in exists:
            if ff_id is None or ff_id == ff.id:
                if ff.status in _ignore_status:
                    continue
                suc = ff.update_tracking(ff_data, message)
                did_update = True
        return exists, did_update
    else:
        return exists, False

def update_shipment_status(order_id, status_data,ff_id=None):
    exists = shopify.Fulfillment.find(order_id=order_id)
    did_update = False
    for ff in exists:
        if ff.status in _ignore_status:
            continue
        if ff_id is None or ff_id == ff.id:
            ff_event = shopify.FulfillmentEvent(
                attributes=status_data, 
                prefix_options={
                    'order_id': order_id,
                    'fulfillment_id': ff.id
                })
            suc = ff_event.save()
            did_update = suc
            
    return did_update


@update_request_handler
def cancel_fulfillment_status(order_id,  ff_id=None):
    exists = shopify.Fulfillment.find(order_id=order_id)
    if exists and len(exists)>0:
        did_update = False
        for ff in exists:
            if ff_id is None or ff_id == ff.id:
                ff.cancel()
                did_update = True
        return exists, did_update

    return None, False

def send_messages():
    pass

@add_request_handler
def add_collection(data):
    c = shopify.CustomCollection(attributes=data)
    suc = c.save()
    return c, suc

@update_request_handler
def udpate_collection(collection_id, data):
    c = shopify.CustomCollection.find(collection_id)
    c._update(data)
    c.save()
    return c

def get_collection_map():
    collections = get_all_items(shopify.CustomCollection, 
                  fields='id,title')
    cmap = {}
    for c in collections:
        cmap[c['title']] = c['id']
    return cmap

@request_handler
def find_collections(**kwargs):
    return shopify.CustomCollection.find(**kwargs)

def get_all_items(item:ActiveResource, **kwargs):
    items = []
    page = get_next_page(item, first_page=True, **kwargs)
    if page:
        items += [r.attributes for r in page]
    # print(page.next_page_url)
    while page and page.next_page_url:
        page = get_next_page(next_url= page.next_page_url)
        if page:
            items += [r.attributes for r in page]
    return items
    


@add_request_handler
def add_image(product_id, data):
    data['product_id'] = product_id
    img = shopify.Image(attributes=data)
    suc = img.save()
    return img, suc


def get_images():
    pass 

@request_handler
def find_metafield(mId=None, resource=None):
    if mId is None:
        m = shopify.Metafield()
        if resource is not None:
            return m.find(**resource)
        else:
            return m.find()
        
    else:
        return shopify.Metafield.find(mId, prefix_options=resource)

@add_request_handler
def add_metafield(data, resource=None):
    m = shopify.Metafield(attributes=data, prefix_options=resource)
    suc = m.save()
    return m, suc




import pandas as pd
import yaml

def read_file_to_df(filename, params:dict=None):
    if params is None: params = {}
    if filename.endswith('.xlsx'):
        return read_excel_file(
                filename,
                sheet_name=params.get('sheet_name', 'Sheet1'))
    elif filename.endswith('.csv'):
        return read_csv_file(filename)


def read_csv_file(filename):
    return pd.read_csv(filename,
                header = 0, 
                dtype=str,
                keep_default_na=False)

def read_excel_file(filename, sheet_name='Sheet1'):
    return pd.read_excel(filename,
                header = 0, 
                dtype=str,
                keep_default_na=False,
                sheet_name=sheet_name)

def read_yml_file(filename):
    data = None 
    with open(filename, 'r') as f:
        data = yaml.safe_load(f)
    return data


# from pypipet.core.sql.model2query import *
# from pypipet.core.sql.query import * 


# def add_supplier_invoice(ctx, session, invoice: dict):
#     if invoice.get('sku') is None:
#         inv_data =  query_select(
#             ctx.get_table_objects().get('inventory'), 
#             session, 
#             filters={'supplier_id': invoice['supplier_id'],
#                     'supplier_item_id': invoice['supplier_item_id']})
#         if len(inv_data) > 0:
#             invoice['sku'] = inv_data[0].sku
        

#     unique_keys = ctx.get_model_keys('supplier_invoice')
#     add_if_not_exist(invoice, 
#                     ctx.get_table_objects().get('supplier_invoice'), 
#                         unique_keys,
#                         session) 
#     return True



# -*- coding: utf-8 -*-

"""
WooCommerce OAuth1.0a Class
"""

__title__ = "woocommerce-oauth"
__version__ = "3.0.0"
__author__ = "Claudio Sanches @ Automattic"
__license__ = "MIT"

from time import time
from random import randint
from hmac import new as HMAC
from hashlib import sha1, sha256
from base64 import b64encode
from collections import OrderedDict
from urllib.parse import urlencode, quote, unquote, parse_qsl, urlparse


class OAuth(object):
    """ API Class """

    def __init__(self, url, consumer_key, consumer_secret, **kwargs):
        self.url = url
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.version = kwargs.get("version", "v3")
        self.method = kwargs.get("method", "GET")
        self.timestamp = kwargs.get("oauth_timestamp", int(time()))

    def get_oauth_url(self):
        """ Returns the URL with OAuth params """
        params = OrderedDict()

        if "?" in self.url:
            url = self.url[:self.url.find("?")]
            for key, value in parse_qsl(urlparse(self.url).query):
                params[key] = value
        else:
            url = self.url

        params["oauth_consumer_key"] = self.consumer_key
        params["oauth_timestamp"] = self.timestamp
        params["oauth_nonce"] = self.generate_nonce()
        params["oauth_signature_method"] = "HMAC-SHA256"
        params["oauth_signature"] = self.generate_oauth_signature(params, url)

        query_string = urlencode(params)

        return f"{url}?{query_string}"

    def generate_oauth_signature(self, params, url):
        """ Generate OAuth Signature """
        if "oauth_signature" in params.keys():
            del params["oauth_signature"]

        base_request_uri = quote(url, "")
        params = self.sorted_params(params)
        params = self.normalize_parameters(params)
        query_params = ["{param_key}%3D{param_value}".format(param_key=key, param_value=value)
                        for key, value in params.items()]

        query_string = "%26".join(query_params)
        string_to_sign = f"{self.method}&{base_request_uri}&{query_string}"

        consumer_secret = str(self.consumer_secret)
        if self.version not in ["v1", "v2"]:
            consumer_secret += "&"

        hash_signature = HMAC(
            consumer_secret.encode(),
            str(string_to_sign).encode(),
            sha256
        ).digest()

        return b64encode(hash_signature).decode("utf-8").replace("\n", "")

    @staticmethod
    def sorted_params(params):
        ordered = OrderedDict()
        base_keys = sorted(set(k.split('[')[0] for k in params.keys()))

        for base in base_keys:
            for key in params.keys():
                if key == base or key.startswith(base + '['):
                    ordered[key] = params[key]

        return ordered

    @staticmethod
    def normalize_parameters(params):
        """ Normalize parameters """
        params = params or {}
        normalized_parameters = OrderedDict()

        def get_value_like_as_php(val):
            """ Prepare value for quote """
            try:
                base = basestring
            except NameError:
                base = (str, bytes)

            if isinstance(val, base):
                return val
            elif isinstance(val, bool):
                return "1" if val else ""
            elif isinstance(val, int):
                return str(val)
            elif isinstance(val, float):
                return str(int(val)) if val % 1 == 0 else str(val)
            else:
                return ""

        for key, value in params.items():
            value = get_value_like_as_php(value)
            key = quote(unquote(str(key))).replace("%", "%25")
            value = quote(unquote(str(value))).replace("%", "%25")
            normalized_parameters[key] = value

        return normalized_parameters

    @staticmethod
    def generate_nonce():
        """ Generate nonce number """
        nonce = ''.join([str(randint(0, 9)) for i in range(8)])
        return HMAC(
            nonce.encode(),
            "secret".encode(),
            sha1
        ).hexdigest()



def map_attr_names(data: dict, field_map: dict):
    "map dict name to alias"
    mapped = {}
    for k, val in data.items():
        mapped_k = field_map.get(k, k)
        mapped[mapped_k] = val
    return mapped

def get_unique_attrs(data: dict, unique_keys:list):
    unique = {}
    for k in unique_keys:
        if data.get(k):
            unique[k] = data[k]
    return unique


"""woocommerce version 5.5.2"""
from .utility import *
import logging
from copy import deepcopy

def exception_handler(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except  TypeError as e:
            logging.debug(e)
    return wrapper



@exception_handler
def wp_parse_order(data:dict, attr_mapping: dict, 
                      shipping_tax_id: int, item_tax_id:int):
    """to do: meta data of other payment mehods"""
    order_data = {}
    if attr_mapping.get('shop_order'):
        field_mapping(data, order_data, attr_mapping['shop_order'])
        if order_data.get('refund') and order_data['refund'] is not None:
            order_data['refund'] = abs(order_data['refund'])
            order_data['order_total'] -= order_data['refund'] 

    # manual input attr value
    order_data['shipping_tax_id'] = shipping_tax_id

    order_data['order_item'] = []
    for item in data['line_items']:
        order_item = {}
        field_mapping(item, order_item, attr_mapping['order_item'])
        order_item['tax_id'] = item_tax_id
        #for variations
        if item.get('variation_id'):
            order_item['destination_product_id'] = str(item['variation_id'])
        order_data['order_item'].append(order_item)
    
    #billing
    order_data['billing_customer'] = {}
    field_mapping(data['billing'], 
                  order_data['billing_customer'], 
                  attr_mapping['customer'])
    order_data['billing_customer']['is_billing'] = True
    #shipping
    if data['shipping']['first_name'].strip() != '' \
    and data['shipping']['last_name'].strip() != '':
        if data['billing']['first_name'] == data['shipping']['first_name'] \
        and data['billing']['last_name'] == data['shipping']['last_name'] \
        and data['billing']['address_1'] == data['shipping']['address_1'] \
        and data['billing']['address_2'] == data['shipping']['address_2'] \
        and data['billing']['postcode'] == data['shipping']['postcode'] :
            order_data['shipping_customer'] = order_data['billing_customer']
            order_data['billing_customer']['is_shipping'] = True
        else:
            order_data['shipping_customer'] = {}
            field_mapping(data['shipping'], order_data['shipping_customer'], 
                            attr_mapping['customer'])
            order_data['shipping_customer']['is_shipping'] = True
    else:
        order_data['shipping_customer'] = order_data['billing_customer']
        order_data['billing_customer']['is_shipping'] = True

    if order_data['shipping_customer'].get('phone') is None \
        or order_data['shipping_customer']['phone'].strip() == '':
        order_data['shipping_customer']['phone'] = order_data['billing_customer']['phone']
    if order_data['shipping_customer'].get('email') is None \
        or order_data['shipping_customer']['email'].strip() == '':
        order_data['shipping_customer']['email'] = order_data['billing_customer']['email']
    
    return order_data



def wp_category_to_taxonomy(cats):
    cat_map = {}
    for c in cats:
        if cat_map.get(c['id']) is None:
            cat_map[c['id']] = {'name': c['name'],  'parent_id': c['parent']}

    for i, c in enumerate(cats):
        parent = None
        full_path = None
        if c['parent'] != 0 and cat_map.get(c['parent']):
            parent =  cat_map[c['parent']]['name']
        
        full_path = _get_full_path(c['id'], cat_map)
        
        cats[i] = {
            'category': c['name'],
            'parent': parent,
            'full_path': full_path
        }


def _get_full_path(category_id, category_map):
    parent_id = category_id
    path_nodes  = []
    while True:
        if parent_id == 0: break
        parent = category_map.get(parent_id, None)
        if parent is None: break
        
        path_nodes.insert(0, parent['name'])
        parent_id = parent['parent_id']
    return '>'.join(path_nodes)

@exception_handler
def parse_product(data:dict, attr_mapping):
    if data is None: return 
    if data.get('parent_id') and type(data['parent_id']) is int:
        #is variations 
        return 0
    if data.get('variations') and len(data['variations']) > 0:
        # is parent
        return 1

    res = {}
    product = {}
    if attr_mapping.get('product'):
        field_mapping(data, product, attr_mapping['product'])
    
    if attr_mapping.get('variation'):
        variation = {}
        field_mapping(data, variation, attr_mapping['variation'])
        if variation.get('sku') is None or variation['sku'].strip() == '':
            variation['sku'] = str(data['id'])
        product['variations'] = [variation]

    res['product'] = product
     
    if attr_mapping.get('destination'):
        dest = {}
        field_mapping(data, dest, attr_mapping['destination'])
        if dest.get('sku') is None or dest['sku'].strip() == '':
            dest['sku'] = str(data['id'])
        res['destinations'] = [dest]
   

    if attr_mapping.get('inventory') and data['manage_stock'] == True:
        inv= {}
        field_mapping(data, inv, attr_mapping['inventory'])
        res['inventory'] = inv 
    return res


@exception_handler
def parse_product_with_variations(data:dict, attr_mapping):
    res = {}
    product = {}
    if attr_mapping.get('product'):
        field_mapping(data, product, attr_mapping['product'])

    default_variation = {}
    field_mapping(data, default_variation, attr_mapping['variation'])
    
    product['variations'] = []
    dests = []
    invs = []
    
    for vari in data['variations']:
        vari_info = deepcopy(default_variation)
        field_mapping(vari, vari_info, attr_mapping['variation_attrs'])
        
        vari_info['images'] = ','.join([vari_info['images'],default_variation['images']])
        # vari_info['description'] = default_variation['description'] + '\n' \
        #                            + vari_info['description']
        if vari.get('sku') is None or vari['sku'].strip() == '':
            vari_info['sku'] = str(data['id'])
        
        product['variations'].append(vari_info)

        dest = {'destination_parent_id': str(data['id'])}
        field_mapping(vari, dest, attr_mapping['destination'])
        if dest.get('sku') is None or dest['sku'].strip() == '':
            dest['sku'] = str(data['id'])
        dests.append(dest)

        inv= {}
        field_mapping(vari, inv, attr_mapping['inventory'])
        invs.append(inv)
   
    res = {
        'product': product,
        'destinations': dests,
        'inventory': invs
    }

    return res



from pypipet.core.operations.order import add_orders_to_db_bulk
from pypipet.core.fileIO.file_saver import write_csv_file
from pypipet.core.operations.inventory import update_inventory_bulk
from pypipet.core.operations.inventory import match_variation_sku_by_upc
from pypipet.core.operations.inventory import update_instock_qty_db
from pypipet.core.operations.inventory import update_instock_front_shop
from pypipet.core.operations.frontshop import update_front_shop_price_bulk
from .utility import *

class Action():
    def __init__(self, name, params:dict=None, **kwargs):
        self.name = name 
        if params is None: params = {}
        self.params = params
        if kwargs.get('func'):
            self.func = kwargs['func']

    def run_external_func(self, data, params:dict):
        return self.func(data, **params)


class ActionMap():
    def __init__(self):
        pass

    @staticmethod
    def mapping(data: list, action):
        if action.name == 'add_orders_to_db':
            add_orders_to_db_bulk(
                action.params.get('table_objs'),
                action.params.get('session'),
                data,
                action.params.get('shop_connector').front_shop_id
            ) 
            return data
        elif action.name == 'model_to_dict':
            if action.params.get('model_name') == 'order':
                data = [order.get_order_info() for order  in data]
            else:
                data = [d.get_all_attrs() for d in data]
            data = [flat_dict(d) for d in data]
            return data
        elif action.name == 'update_inventory_to_db':
            update_inventory_bulk(
                action.params.get('table_objs'),
                action.params.get('session'),
                data
            )
            return data
        elif action.name == 'match_inventory_sku':
            match_variation_sku_by_upc(
                action.params.get('table_obj'),
                action.params.get('session'),
                data
            )
            return data
        elif action.name == 'update_instock_qty_db':
            update_instock_qty_db(
                action.params.get('table_objs'),
                action.params.get('session'),
                batch_size=action.params.get('batch_size'),
                params=action.params.get('params', {})
            )
            return data
        elif action.name == 'update_instock_qty_shop':
            update_instock_front_shop(
                action.params.get('table_objs'),
                action.params.get('session'),
                action.params.get('shop_connector'),
                batch_size=action.params.get('batch_size'),
                params=action.params.get('params', {})
            )
            return data
        # elif action.name == 'sync_order_item_from_shop':
        #     update_order_items_from_shop(
        #         action.params.get('table_objs'),
        #         action.params.get('session'),
        #         action.params.get('shop_connector'),
        #         data)
            
        #     return data
        elif action.name == 'update_front_shop_price':
            update_front_shop_price_bulk(
                action.params.get('table_objs'),
                action.params.get('session'),
                action.params.get('shop_connector'),
                data)
            
            return data

        elif action.name == 'save_to_csv':
            write_csv_file(data, 
                           action.params.get('save_to', 
                                            'pp_save_to_csv.csv'))
            return data
    
    


import click

def col2dict(sku, row, default_brand):
    d = {'product': {'variations': {}}, 'destination': {}}
    for k, val in dict(row).items():
        # print(k)
        if val.strip() == '': continue
        table, col = k.split('.')
        if table == 'product':
            d[table][col] = val
        elif table == 'destination':
            if col == 'price': val = float(val)
            if col == 'destination_product_id': val = str(val)
            if col == 'discount': val = float(val)
            if col == 'available': val = not val.lower() == 'false'
            
            d[table][col] = val
        elif table == 'variation':
            if col == 'upc': val = str(val)
            if col == 'in_stock': val = int(val)
            if val in ('True', 'False'):
                val = val == True
            d['product']['variations'][col] = val 
    
    if d['product'].get('brand') is None:
        d['product']['brand'] = default_brand
    if d['product'].get('short_description') is None \
            or d['product'].get('short_description') == '':
        d['product']['short_description']= \
                    d['product']['product_name']

    d['product']['variations'] = [d['product']['variations']]
    d['destination']['sku'] = sku
    return d


def col2dict_update(sku, row):
    d = {'product': {}, 'variation': {}, 'destination': {}}
    for k, val in dict(row).items():
        if val.strip() == '': continue
        table, col = k.split('.')
        if table == 'product':
            if col == 'id': val = int(val)
            d[table][col] = val
        elif table == 'destination':
            if col == 'price': val = float(val)
            if col == 'destination_product_id': val = str(val)
            if col == 'discount': val = float(val)
            if col == 'available': val = not val.lower() == 'false'
            
            d[table][col] = val
        elif table == 'variation':
            if col == 'upc': val = str(val)
            if col == 'in_stock': val = int(val)
            d['variation'][col] = val 
    
    if len(d['destination']) > 0:
        d['destination']['sku'] = sku
    return d

def map_product_sku_for_launch(df):
    res = []
    for i, row in df.iterrows():
        row = dict(row)
        price = None 
        if _has_value(row.get('destination.price'), data_type='float'):
            price =float(row['destination.price'])
        elif _has_value(row.get('price'), data_type='float'):
            price =float(row['price'])
        if price is None:
            click.secho('missing price in row {}'.format(i+1),
                   fg='yellow')
            continue

        sku = None 
        if _has_value(row.get('variation.sku')):
            sku = row['variation.sku']
        elif _has_value(row.get('sku')):
            sku = row['sku']

        if sku is None:
            click.secho('missing sku in row {}'.format(i+1),
                   fg='yellow')
            continue
        res.append({'sku': sku, 'price': price})
    return res

        

def _has_value(d, data_type=None):
    if d is None: return False
    if d.strip() == '': return False
    if data_type == 'float':
        try:
            d = float(d)
        except:
            return False 
    return True



import logging
logger = logging.getLogger('__default__')

def field_mapping(data_from:dict, data_to:dict, attr_map:dict):
    for k, val in attr_map.items():
        if type(val) is list:
            if len(data_from[k]) == 0: continue
            fields = _map_from_list(data_from[k], val)  
            data_to.update(fields) 
        elif type(val) is dict and val.get('map_to') is not None:
            mapped_value =_map_from_field(data_from[k], val)
            if mapped_value is not None:
                data_to[val['map_to']] = mapped_value
        elif type(val) is str:
            if data_from.get(k) is not None:
                data_to[val] = data_from.get(k) 



def _map_from_field(field_value, attr_map:dict):
    if attr_map.get('data_type') is not None:
        try:
            if attr_map['data_type'] == 'string':
                field_value = str(field_value)
            elif attr_map['data_type'] == 'float':
                field_value = float(field_value)
            elif attr_map['data_type'] == 'integer':
                field_value = int(field_value)
            elif attr_map['data_type'] == 'boolean':
                if attr_map.get('data_value') is not None:
                    field_value = field_value ==  attr_map['data_value']
                field_value = bool(field_value)
        except Exception as e:
            logger.debug(e)
            logger.debug([field_value])
            field_value = None 
    return field_value

def _map_from_list(data: list, attr_map_list:list):
    """
    attr_map_list [{map_from: xxx, map_to: xxx}]
    """
    res = {}
    for attr_map in attr_map_list:

        if attr_map.get('index') is not None:
            res[attr_map['map_to']] = \
                        data[attr_map['index']][attr_map['map_from']]
        elif attr_map.get('merge') is not None:
            res[attr_map['map_to']] = _merge_items(data, attr_map['map_from'], sep=',')
        elif attr_map.get('sum') is not None:
            res[attr_map['map_to']] = _sum_items(data, attr_map['map_from'])
        else:
            if  attr_map.get('condition') is not None:
                cond = list(attr_map['condition'].items())[0]
                value_dict = _search_from_list(data, cond[0], cond[1])
                if value_dict is None: continue

                map_value = value_dict.get(attr_map['map_from'])
                if map_value is None: continue
                elif type(map_value) is list:
                    map_value = '-'.join(map_value) 
                
                res[attr_map['map_to']] = map_value
            
            else:
                res[attr_map['map_to']] = data[0].get(attr_map['map_from'])
        
        if attr_map.get('data_type'):
            res[attr_map['map_to']] = _map_data_type(res[attr_map['map_to']], 
                                                     attr_map['data_type'])

    return res


def _map_data_type(val, data_type):
    try:
        if data_type == 'string':
            return str(val)
        elif data_type == 'float':
            return float(val)
        elif data_type == 'integer':
            return int(val)
        elif data_type == 'boolean':
            return bool(val)
    except Exception as e:
        logger.debug(e)
        return val 

def _search_from_list(data:[dict], key, value):
    for d in data:
        if d.get(key) == value:
            return d 
    return None

def _merge_items(data_list: [dict], key:str, sep='-'):
    items = []
    for d in data_list:
        if d.get(key) is not None:
            items.append(d[key])
    return sep.join(items)


def _sum_items(data_list: [dict], key:str):
    val = 0
    for d in data_list:
        if d.get(key) is not None:
            try:
                val += float(d[key])
            except:
                pass 
            
    return val



"""Defines helpers related to the system database."""

import logging
import time

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import text

from .errors import DialectNotSupportedError
from .sql.init_db import *
from .sql.query_interface import insert_list_raw


def project_engine(db_setting):
    """Creates and register a SQLAlchemy engine"""
    engine_uri = dialect_engine_uri(db_setting['db_setting']['db_type'],
                       db_setting['db_setting']['db_conn'])
    logging.debug(f"Creating engine {engine_uri}")
    engine = create_engine(engine_uri, pool_pre_ping=True)

    check_db_connection(
        engine,
        max_retries=db_setting['db_setting'].get("database_max_retries"),
        retry_timeout=db_setting['db_setting'].get("database_retry_timeout"),
    )
    
    if db_setting['db_setting']['db_type'] == 'postgres' \
    and db_setting['db_setting']['db_conn'].get('schema'):
        session_maker = sessionmaker(bind=engine)
        session = session_maker()

        add_schema(session, db_setting['db_setting']['db_type'], 
                    db_setting['db_setting']['db_conn']['schema'])
        set_schema(session, db_setting['db_setting']['db_type'], 
                    db_setting['db_setting']['db_conn']['schema'])
        session.close()
        
        # drop_all_tables(db_setting['db_setting']['tables'], engine,
        #                 schema=db_setting['db_setting']['db_conn'].get('schema'))
        
    
    init_tables(engine, db_setting['db_setting']['tables'], 
               schema=db_setting['db_setting']['db_conn'].get('schema'))

    return engine

def dialect_engine_uri(dialect, params={}):
        # engine_uri according to the db type
        db_suffix = lambda s: str(Path(s).with_suffix(".db"))

        dialect_templates = {
            "postgres": lambda params: "postgresql://{user}:{password}@{host}:{port}/{dbname}".format(
                **params
            ),
            "snowflake": lambda params: "snowflake://{username}:{password}@{account}/{database}/{schema}?warehouse={warehouse}&role={role}".format(
                **params
            ),
            "sqlite": lambda params: "sqlite:///{database}".format(
                database=db_suffix(params.pop("database")), **params
            ),
        }

        try:
            url = dialect_templates[dialect](
                params
            )

            return url
        except KeyError:
            raise DialectNotSupportedError(dialect)

def check_db_connection(engine, max_retries, retry_timeout):  
    """Check if the database is available the first time a project's engine is created."""
    attempt = 0
    while True:
        try:
            engine.connect()
        except OperationalError:
            if attempt == max_retries:
                logging.error(
                    "Could not connect to the Database. Max retries exceeded."
                )
                raise
            attempt += 1
            logging.info(
                f"DB connection failed. Will retry after {retry_timeout}s. Attempt {attempt}/{max_retries}"
            )
            time.sleep(retry_timeout)
        else:
            break


def import_data(table_objs, session, input_tables: dict):
    for tablename, data in  input_tables.items():
        insert_list_raw( table_objs.get(tablename), 
                         session,
                          data)


def set_schema(session, dialect, schema_name):
    if schema_name is None:
        return
    if dialect == 'oracle':
        session.execute(text(f'ALTER SESSION SET CURRENT_SCHEMA={schema_name}'))
    elif dialect == 'postgres':
        session.execute(text(f'SET search_path={schema_name}'))

def add_schema(session, dialect, schema_name):
    if schema_name is None:
        return 
    create_schema = None
    if dialect == 'postgres':
        create_schema = text(f"CREATE SCHEMA IF NOT EXISTS {schema_name}")
    
    if create_schema is not None:
        # print(create_schema)
        session.execute(create_schema)
        session.commit()
            



import click 
from .cli import cli
from pypipet.core.operations.utility import get_front_shop_id
from pypipet.core.operations.fulfillment import get_shipping_info
from pypipet.core.operations.fulfillment import add_fulfillment_bulk
from pypipet.core.operations.fulfillment import add_fulfillment
from pypipet.core.operations.fulfillment import update_tracking
from pypipet.core.operations.fulfillment import update_fulfillment_to_front_shop
from pypipet.core.sql.query_interface import search_exist
from pypipet.core.fileIO.file_loader import read_csv_file
from .utility import col2dict_update
from pprint import pprint


@cli.command()
@click.argument("action", type=click.Choice(['add', 'edit', 'show']))
@click.option("--shop", 
                help="frontshop name (in setting.shops)")
@click.option("--tracking",
                help="tracking code")
@click.option("--provider",type=click.Choice(['canada_post','usps',
             'ups', 'fedex', 'dhl']), help="shipping service provider")
@click.option("--order-id", help="fronshop order id")
@click.option("-f", "--filename",
                help=" edit tracking info in bulk")
@click.option('--message', is_flag=True, 
                        help="send message to customers.")
@click.pass_context
def fulfillment(ctx, action, shop, tracking, provider, order_id, filename, message):
    """
    add/edit shipping tracking (by skus or by file)
    """
    click.echo('{} fulfillment at {}... send mesage \
     to customer: {}'.format(action, shop, message))

    project = ctx.obj['project']
    config = ctx.obj["config"]
    if config is None:
        click.secho('create project first',
                   fg='yellow')
        return

    project.initialize_project(config)
    table_classes = project.get_table_objects()
    sessionmaker = project.get_session_maker()
    session = sessionmaker()
    
    shop_conn = project.get_shop_connector(shop)
    if shop_conn is None:
        click.secho('shop setting not found: {}'.format(shop),
                   fg='yellow')
        session.close()
        return 
    
    get_front_shop_id(table_classes, session, shop_conn)


    if action == 'add':
        _add_tracking(session, table_classes, shop_conn, order_id, 
                                    tracking, provider, filename, message=message)
    elif action == 'edit':
        _update_fulfillment(session, table_classes, shop_conn, order_id, 
                                          tracking, provider, filename,  message=message)
    elif action == 'show':
        shipment = None
        if tracking is not None:
            shipment = get_shipping_info(table_classes, session,
                                     {'tracking_id': tracking})
        else:
            shipment = get_shipping_info(table_classes, session,
                                     {'destination_order_id': order_id,
                                      'front_shop_id': shop_conn.front_shop_id})
        click.echo(shipment)
    else:
        click.secho('action not supported: {}'.format(action),
                fg='yellow')
    
    session.close()


def _add_tracking(session, table_classes, shop_conn, order_id, 
                         tracking, provider, filename, message=False):
    if filename is None:
        if order_id is None or tracking is None or provider is None:
            click.echo('invalid tracking info')
            return 
        
        data = {
            'tracking_id': tracking,
            'provider': provider,
            'destination_order_id': order_id
        }
        _edit_fulfilment(session, table_classes, data, 
                           shop_conn.front_shop_id,order_id, action=1, 
                           shop_conn=shop_conn, message=message)
        
    else:
        _add_tracking_by_file(session, table_classes, shop_conn, filename, message=message) 

def _add_tracking_by_file(session, table_classes, shop_conn, filename, message=False):
    data = _get_data_from_file(filename)
    for d in data:
        if d.get('front_shop_id') is None or  d['front_shop_id'] == '':
            d['front_shop_id'] = shop_conn.front_shop_id

        _edit_fulfilment(session, table_classes, {
                'tracking_id': d['tracking_id'],
                'provider': d['provider'],
                'destination_order_id': d['destination_order_id']
            }, d['front_shop_id'],d['destination_order_id'], action=1,
            shop_conn=shop_conn, message=message)


def _update_fulfillment(session, table_classes, shop_conn, order_id, 
                            tracking, provider, filename, message=False):
    if filename is None:
        if order_id is None or (tracking is None and provider is None):
            click.echo('invalid tracking info')
            return 
        
        data = {
            'tracking_id': tracking,
            'provider': provider,
            'destination_order_id': order_id
        }
        _edit_fulfilment(session, table_classes, data, 
                shop_conn.front_shop_id,order_id, action=0, 
                shop_conn=shop_conn, message=message)
    else:
        _update_tracking_by_file(session, table_classes, shop_conn, 
                         filename, message=message) 


def _update_tracking_by_file(session, table_classes, shop_conn, filename,
                                                   message=False):
    data = _get_data_from_file(filename)
    for d in data:
        if d.get('front_shop_id') is None or  d['front_shop_id'] == '':
            d['front_shop_id'] = shop_conn.front_shop_id
        
        _edit_fulfilment(session, table_classes, {
                'tracking_id': d['tracking_id'],
                'provider': d['provider'],
            }, d['front_shop_id'],d['destination_order_id'], action=0,
            shop_conn=shop_conn, message=message)

def _edit_fulfilment(session, table_classes, fulfillment_info, 
                 front_shop_id,order_id, action=0, shop_conn=None,message=False):
    """action 0: update
       actiion 1: add"""
    
    if action == 1:
        fulfillment_info.update({
            'front_shop_id': front_shop_id
            })
        click.echo('adding tracking {}'.format(fulfillment_info))
        res = add_fulfillment(table_classes, session, fulfillment_info)

        if message and res is not None:
            click.echo('send message to customer')
            update_fulfillment_to_front_shop(table_classes, session, 
                shop_conn, {}, tracking_info=fulfillment_info)

    elif action == 0:
        click.echo('update tracking {}'.format(fulfillment_info))
        order_info = _get_order(session, table_classes.get('shop_order'), 
                                order_id, front_shop_id)
        update_tracking(table_classes, session, fulfillment_info,
                                 {'shop_order_id':order_info['id']})
    
        if message:
            click.echo('send message to customer')
            update_fulfillment_to_front_shop(table_classes, session, 
                shop_conn, {}, tracking_info=fulfillment_info)

    return fulfillment_info

def _get_order(session, table_obj, order_id, front_shop_id):
    order = search_exist(table_obj, session, {
        'destination_order_id': order_id,
        'front_shop_id': front_shop_id
    })
    if len(order) > 0:
        return _object2dict(order[0])


def _get_data_from_file(filename):
    data = read_csv_file(filename)
    ff = []

    #validation
    for i, row in data.iterrows():
        if row['destination_order_id'].strip() == '':
            click.secho('row {} missing order id from destination shop'\
                  .format(i+1), 
                        fg='yellow')
            continue 
        if row['tracking_id'].strip() == '':
            click.secho('row {} missing tracking_id from destination shop'\
                  .format(i+1), 
                        fg='yellow')
            continue 
        ff.append(dict(row))

    return ff
    

def _object2dict(table_obj):
    d = {}
    for k in table_obj.__table__.columns.keys():
        d[k] = table_obj.__dict__.get(k, None)
    return d


"""Base Error classes."""
import functools
import io
import logging

def error(error_cls):
    class Error(error_cls):
        """Aggregate multiple sub-exceptions."""

        def __init__(self, exceptions):
            self.exceptions = exceptions

        def __str__(self):
            return "\n".join(str(e) for e in self.exceptions)

    if error_cls != Exception:
        error_cls.Aggregate = Error

    return error_cls


CustomerizedError = error(Exception)

class DialectNotSupportedError(Exception):
    def exit_code(self):
        return 1



import pypipet.core.shop_conn.wc as wc
import pypipet.core.transform.wc_to_model as wc_to_model
import pypipet.core.transform.model_to_wc as model_to_wc


class ShopConnector(object):
    def __init__(self, shop_name, shop_type, **kwargs):
        self.shop_name = shop_name
        self.shop_type = shop_type
        self.shop_api = kwargs.get('api', None)
        self.front_shop_id = kwargs.get('front_shop_id', None)
        self.field_mapping = kwargs.get('field_mapping', None)
        

    def set_shop_api(self, api):
        self.shop_api = api 

    def set_front_shop_id(self, shop_id):
        self.front_shop_id = shop_id

    def set_field_mapping(self, mapping):
        self.field_mapping = mapping

    def get_shop_info(self):
        return {
            'shop_name': self.shop_name,
            'shop_type': self.shop_type,
            'front_shop_id': self.front_shop_id
        }


    def add_product_to_shop(self,product_info: dict, **kwargs):
        pass

    def add_product_to_shop_bulk(self, products:list,  **kwargs):
        pass


    def update_shop_product(self, data:dict, product_id:str, **kwargs ):
        pass

    def update_shop_product_batch(self, data:list, **kwargs):
        pass

    def sync_shop_orders(self,**kwargs):
        pass

    def update_order_status_at_shop(self, destination_order_id: str, 
                                   status: str,**kwargs):
        pass

    def update_order_at_shop(self, destination_order_id: str, 
                             data, **kwargs):
        pass

    def update_order_item_at_shop(self, destination_order_id: str, 
                             items: list, **kwargs):
        pass

    def get_order_at_shop(self, **kwargs):
        pass

        
    def get_products_at_shop(self,**kwargs):
        pass
                                    
    def get_product_at_shop(self, product_id: str, **kwargs):
        pass
    
    def get_category_taxonomy(self, **kwargs):
        pass
    
    def update_customer_at_shop(self,destination_order_id: str, 
                             data: dict, **kwargs):
        pass 

    def add_tracking(self,destination_order_id: str, 
                            message: dict, **kwargs):
        pass

    def send_message(self, destination_order_id: str, 
                            message: str, **kwargs):
        pass





import logging, sys
from logging.handlers import TimedRotatingFileHandler

def get_console_handler(formatter):
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    return console_handler

def get_file_handler(formatter, log_file):
    file_handler = TimedRotatingFileHandler(log_file)
    file_handler.setFormatter(formatter)
    return file_handler


# -*- coding: utf-8 -*-

"""
woocommerce
~~~~~~~~~~~~~~~
A Python wrapper for WooCommerce API.

:copyright: (c) 2019 by Automattic.
:license: MIT, see LICENSE for details.
"""

__title__ = "woocommerce"
__version__ = "3.0.0"
__author__ = "Claudio Sanches @ Automattic"
__license__ = "MIT"



from copy import deepcopy
from math import prod
from pprint import pprint

def add_metafields(attributes: dict, attr_list, option_attrs=None, 
                namespace='variant'):
    # shopify: max two options in option_attrs
    # add more to metafields
    meta = []
    for attr, value in attributes.items():
        if attr not in attr_list: continue
        if option_attrs and attr in option_attrs:
            pass
        else:
            meta.append({
                'key': attr,
                'value': str(value),
                "type": "single_line_text_field",
                'namespace': namespace,
            })
    if len(meta) > 0:
        return meta

def variant_parser(variation, option_attrs=None):
    if variation:
            for j, attr in enumerate(option_attrs):
                if variation.get(attr) is not None:
                    variation[f'option{j+1}'] = variation[attr]

            meta = add_metafields(variation, option_attrs=option_attrs)
            if meta:
                variation['metafields'] = meta
            
            # del variation['attributes']
    if variation.get('images'):
        variation['images'] = transform_imgs(variation['images'])

def product_parser(product, attr_list, option_attrs, add_new=True, **kwargs):
    if product.get('created_at'): del product['created_at']
    if product.get('updated_at'): del product['updated_at']
    if product.get('id'): del product['id']
    if product.get('product_id'): del product['product_id']
    # if product.get('short_description'):del product['short_description']
    if product.get('short_description'):
        product['body_html'] = product['short_description']
        del product['short_description']

    if product.get('brand') is not None:
        product['vendor'] = product['brand']

    if kwargs.get('weight_unit'):
        product['weight_unit'] = kwargs['weight_unit']
    

    if add_new and product.get('category') and product.get('product_type') is None:
        product['product_type'] = product['category']

    imgs, variation_imgs = _parser_img_mapping(product)
    if imgs and len(imgs) > 0:
        product['images'] = imgs

    variants = product.get('variations')
    if variants is None:
        #if it is simple product
        _clean_null(product)
        variants = [deepcopy(product)]

    exist_options = set()

    for i, vari in enumerate(variants):   
        if vari.get('created_at'): del vari['created_at']
        if vari.get('updated_at'): del vari['updated_at']
        if vari.get('product_name'): del vari['product_name']
        if vari.get('body_html'): del vari['body_html']
        if vari.get('id'): del vari['id']
        if vari.get('product_id'): del vari['product_id']
        # if vari.get('description'): 
        #     if product.get('body_html') is None:
        #         product['body_html'] = vari['description']
        #     del vari['description']
            
        if vari.get('upc'):
            vari['barcode'] = vari['upc']
        if vari.get('ean'):
            vari['barcode'] = vari['ean']
        
        if option_attrs:
            for j, attr in enumerate(option_attrs):
                if vari.get(attr) is not None:
                    variants[i][f'option{j+1}'] = vari[attr]
                    exist_options.add(attr)

        meta = add_metafields(vari, attr_list, option_attrs=option_attrs)
        if meta:
            variants[i]['metafields'] = meta

        if add_new:
            variants[i]['inventory_management'] = kwargs.get('inventory_management',
                                                            'shopify')
            variants[i]['fulfillment_service'] = kwargs.get('fulfillment_service',
                                                            'manual')
        if vari.get('discount'):
            assert vari['price']
            vari['compare_at_price'] = vari['price'] 
            vari['price'] = vari['price'] - vari['discount']

        #existing parent
        if kwargs.get('parent_id') is None:
            if vari.get('images'): del vari['images']
        elif vari.get('images'):
            variants[i]['images'] = transform_imgs(vari['images'])

        
    product['variants'] = variants
    if product.get('product_name'):
        product['title'] = product['product_name']

    if product.get('variations'):
        del product['variations']
    # else:
    #     meta_product = add_metafields(product, attr_list)
    #     if meta_product:
    #         product['metafields'] = meta_product

    
    product['variants'] = variants 

    if add_new and option_attrs:
        product['options'] = []
        for attr in option_attrs:
            if attr in exist_options: 
                product['options'].append({'name': attr})

    
    return variation_imgs

def transform_customer(data):
    if data.get('postcode') is not None:
        data['zip'] = data['postcode']
    if data.get('state') is not None:
        data['state'] = data['province']
    if data.get('email') is not None:
        data['email'] = data['email'].lower()
    if data.get('updated_at') is not None:
        del data['updated_at']
    if data.get('created_at') is not None:
        del data['created_at']

def transform_tracking(data, message=True, add_new=True):
    if data:
        if add_new:
            ff_info = {'message': message}
            if data.get('tracking_id'):
                ff_info['tracking_numbers'] = data['tracking_id'].split(',')
            if data.get('tracking_urls'):
                ff_info['tracking_urls'] = data['tracking_urls'].split(',')
            if data.get('provider'):
                ff_info['tracking_company'] = data['provider']
            return ff_info 
        else:
            ff_info = {}
            if data.get('tracking_id'):
                ff_info['number'] = data['tracking_id']
            if data.get('tracking_urls'):
                ff_info['url'] = data['tracking_urls']
            if data.get('provider'):
                ff_info['company'] = data['provider']
            return ff_info    

def transform_order_item(items:dict):
    for variant_id, value in items.items():
        if value.get('destination_product_id'):
            items[variant_id]['variant_id'] = value['destination_product_id']
            del items[variant_id]['destination_product_id']
        if value.get('destination_parent_id'):
            items[variant_id]['product_id'] = value['destination_parent_id']
            del items[variant_id]['destination_parent_id']
        if value.get('sku'):
            items[variant_id]['sku'] = value['sku']
        
        if value.get('order_qty'):
            items[variant_id]['quantity'] = value['order_qty']
            del items[variant_id]['order_qty']
        elif value.get('qty'):
            items[variant_id]['quantity'] = value['qty']
            del items[variant_id]['qty']

    

def _clean_null(data: dict):
    keys = data.keys()
    deleted_keys = []
    for k  in keys:
        if data[k] is None:
            deleted_keys.append(k)

    for k in deleted_keys:
        del data[k]

def transform_imgs(img_str, sep=','):
    return [{'src': img} 
           for img in img_str.split(sep)]

def _parser_img_mapping(product):
    imgs = []
    variation_imgs = {}
    if product.get('variations'):
        for vari in product['variations']:
            if vari.get('images'):
                variation_imgs[vari['sku']] = len(imgs)
                imgs += transform_imgs(vari['images'])
    elif product.get('images'):
        imgs += transform_imgs(product['images'])
    return imgs, variation_imgs


from pypipet.core.model.user import User
from .utility import _object2dict
from pypipet.core.sql.query_interface import (
    add_to_db, search_exist, update_data)

import logging
logger = logging.getLogger('__default__')

def register_new_user(table_objs, session, user):
    if user is dict:
        user = User(user)
    
    return add_to_db(table_objs.get(user.__table_name__), session, user) 


def find_users(table_obj, session, params):
    res = search_exist(table_obj, session, params)
    if res and len(res) >0:
        return [_object2dict(r) for r in res]


def change_password():
    pass


"""shopify version 2022-01"""
from builtins import isinstance
from .utility import *
import logging

def exception_handler(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except  TypeError as e:
            logging.debug(e)
    return wrapper



@exception_handler
def parse_order(data, attr_mapping,shipping_tax_id=None, item_tax_id=None, **kwargs):
    if data is None: return 
    order_data = {}
    if attr_mapping.get('shop_order'):
        field_mapping(data, order_data, attr_mapping['shop_order'])
        if data.get('cancelled_at') is not None:
            order_data['status'] = 'cancelled'
    if shipping_tax_id:
        order_data['shipping_tax_id'] = shipping_tax_id

    order_data['order_item'] = []
    for item in data['line_items']:
        order_item = {}
        field_mapping(item, order_item, attr_mapping['order_item'])
        if item_tax_id:
            order_item['tax_id'] = item_tax_id
        order_data['order_item'].append(order_item)

    order_data['billing_customer'] ={'email': data.get('email'),
                                    'phone': data.get('phone'),
                                    'is_billing': True}
    if data.get('billing_address'):
        field_mapping(data['billing_address'], 
                  order_data['billing_customer'], 
                  attr_mapping['customer'])
    
    if data.get('shipping_address'):
        if data.get('billing_address') and data['shipping_address']['first_name'].strip() != '' \
    and data['shipping_address']['last_name'].strip() != '':
            if data['billing_address']['first_name'] == data['shipping_address']['first_name'] \
            and data['billing_address']['last_name'] == data['shipping_address']['last_name'] \
            and data['billing_address']['address1'] == data['shipping_address']['address1'] \
            and data['billing_address']['address2'] == data['shipping_address']['address2'] \
            and data['billing_address']['zip'] == data['shipping_address']['zip'] :
                order_data['shipping_customer'] = order_data['billing_customer']
                order_data['billing_customer']['is_shipping'] = True
            else:
                order_data['shipping_customer'] = {}
                field_mapping(data['shipping_address'], order_data['shipping_customer'], 
                                attr_mapping['customer'])
                order_data['shipping_customer']['is_shipping'] = True
        else:
            order_data['shipping_customer'] = order_data['billing_customer']
            order_data['billing_customer']['is_shipping'] = True

    if data.get('fulfillments') and len(data['fulfillments']) > 0 and attr_mapping.get('fulfillment'):
        order_data['fulfillment'] = []
        for ff_info in data['fulfillments']:
            if ff_info.get('status') != 'success': continue
            ff_info['status'] = 'shipped'  # mapped to shipped
            ff = {'destination_order_id': str(data['id'])}
            field_mapping(ff_info, 
                  ff, 
                  attr_mapping['fulfillment'])
            if ff.get('tracking_id') and isinstance(ff['tracking_id'], list):
                ff['tracking_id'] = ','.join(ff['tracking_id'])
            order_data['fulfillment'].append(ff)
    return order_data

@exception_handler
def parse_product(data: dict, attr_mapping):
    if data is None: return 

    res = {}
    product = {}
    options = []
    if attr_mapping.get('product'):
        field_mapping(data, product, attr_mapping['product'])
        if product.get('category') is None:
            product['category'] = 'undefined'
        options = [opt['name'] for opt in data['options']]
    
    if attr_mapping.get('variation'):
        product['variations'] = []
        for vari in data.get('variants'):
            variation = {}
            vari.update({ 'images': data['images'],
                          'meta': vari.get('meta',[]) + data.get('meta',[])})
            field_mapping(vari, variation, attr_mapping['variation'])
            if variation.get('sku') is None or variation['sku'].strip() == '':
                variation['sku'] = str(vari['id'])
            # variation['description'] = f"{data['body_html']} - {variation['description']}"
            variation['title'] = f"{product['product_name']} ({variation['title']})"
            #options: shopfy has max 3 options
            for i in range(1,4):
                if vari.get(f'option{i}') is not None and i <= len(options):
                    variation[options[i-1]] = vari[f'option{i}']

            product['variations'].append(variation)

    res['product'] = product
     
    if attr_mapping.get('destination'):
        res['destinations'] = []
        for vari in data.get('variants'):
            dest = {}
            vari.update({'parent_id': data['id']})
            field_mapping(vari, dest, attr_mapping['destination'])
            if dest.get('sku') is None or dest['sku'].strip() == '':
                dest['sku'] = str(vari['id'])
            if dest.get('discount'):
                dest['discount']  -=  dest['price']
                if dest['discount'] > 0:
                    dest['price'] += dest['discount'] 
            res['destinations'].append(dest)
   

    return res






import binascii,os
import shopify
from urllib.parse import urlparse
from urllib.parse import parse_qs

def get_auth_url(API_KEY, API_SECRET, shop_url,  scopes, redirect_uri, **kwargs):
    shopify.Session.setup(api_key=API_KEY, secret=API_SECRET)
    state = kwargs.get('state', binascii.b2a_hex(os.urandom(15)).decode("utf-8"))
    api_version = kwargs.get('api_version','2020-10')

    session = shopify.Session(shop_url, api_version)
    return session.create_permission_url(scopes, redirect_uri, state)

def get_acess_token(shop_url, target_url, **kwargs):
    api_version = kwargs.get('api_version','2020-10')
    session = shopify.Session(shop_url, api_version)
    params = _parse_target_url(target_url)
    return session.request_token(params) 

def start_session(private_app=True, **kwargs):
    api_version = kwargs.get('api_version','2020-10')
    if private_app:
        return shopify.Session(kwargs['shop_url'], api_version, 
                      kwargs['private_app_password'])
    else:
        return shopify.Session(kwargs['shop_url'], api_version, 
                    kwargs['access_token'])

def _parse_target_url(url):
    parsed_url = urlparse(url)
    params = parse_qs(parsed_url.query)
    for key, value in params.items():
        params[key] = value[0]
    return params



from dataclasses import field
import pypipet.core.shop_conn.spf as spf
from pypipet.core.sql.query_interface import add_new_product
import pypipet.core.transform.spf_to_model as spf_to_model
import pypipet.core.transform.model_to_spf as model_to_spf
from .shop_connector import ShopConnector
import shopify
import logging
logger = logging.getLogger('__default__')

class SPFShopConnector(ShopConnector):
    def __init__(self, shop_name, shop_type, **kwargs):
        super().__init__(shop_name, shop_type, **kwargs)
        self.batch_size = kwargs.get('batch_size', 5)
        self.collections = None
        self.product_ids = None
        self.order_ids = None

    def get_all_collections(self, **kwargs):
        has_active_session = kwargs.get('has_active_session', False)
        if has_active_session == False:
            shopify.ShopifyResource.activate_session(self.shop_api)
        self.collections = spf.get_collection_map()
        if has_active_session == False:
            shopify.ShopifyResource.clear_session()

    def add_collection(self, name,**kwargs):
        has_active_session = kwargs.get('has_active_session', False)
        if has_active_session == False:
            shopify.ShopifyResource.activate_session(self.shop_api)
        
        if self.collections is None: self.get_all_collections(has_active_session=True)
        collection_id = self.collections.get(name)
        if collection_id is None:
            collection_id, suc = spf.add_collection({'title': name})
            if not suc: logger.debug(collection_id)
            collection_id = collection_id['id']
            self.collections[name] = collection_id
            
        if has_active_session == False:
            shopify.ShopifyResource.clear_session()
        return collection_id

    def add_product_to_shop(self,product_info: dict, **kwargs):
        has_active_session = kwargs.get('has_active_session', False)
        if has_active_session == False:
            shopify.ShopifyResource.activate_session(self.shop_api)
        attr_list = kwargs.get('attr_list', ['size', 'color'])

        is_simple_product = product_info.get('variations') is None
        
        variation_imgs = model_to_spf.product_parser(product_info, 
                     attr_list,
                     kwargs.get('variation_attrs'),
                     parent_id=kwargs.get('parent_id'),
                     )
        msg = None 
        suc = False
        if kwargs.get('parent_id') is None:
            msg, suc = spf.add_product(product_info, variant_imgs=variation_imgs)
        else:
            msg = {'id': kwargs['parent_id'], 'variants': []}
            for vari in product_info['variants']:
                msg_variant, suc = spf.add_variation(msg['id'], vari)
                if suc == True: msg['variants'].append(msg_variant)
            
        if suc == True:
            if kwargs.get('parent_id') is None and product_info.get('category') is not None:
                collection_id = self.add_collection(product_info['category'],
                                                   has_active_session=True)
                if collection_id:
                    spf.add_product_to_collection(msg['id'], collection_id)
            # if kwargs.get('update_inventory'):
            #     for i,vari in enumerate(product_info['variants']):
            #         qty = vari.get('in_stock', product_info.get('in_stock'))
            #         inventory_item_id = msg['variants][i]['inventory_item_id']
            #         if qty:
            #             self.update_shop_product({'inventory_item_id': inventory_item_id, 
            #                                    'in_stock': qty}, 
            #                                   vari['id'],
            #                                   has_active_session=True)    
        
        if has_active_session == False:
            shopify.ShopifyResource.clear_session()

        if suc == True:
            return self._parse_product(msg, is_simple_product=is_simple_product)

        logger.debug(msg)

    
    def add_product_to_shop_bulk(self, products:list,  **kwargs):
        shopify.ShopifyResource.activate_session(self.shop_api)
        for product in products:
            self.add_product_to_shop(product, has_active_session=True, **kwargs)
        shopify.ShopifyResource.clear_session()

    def update_shop_product(self, data:dict, product_id:str, **kwargs ):
        has_active_session = kwargs.get('has_active_session', False)
        if has_active_session == False:
            shopify.ShopifyResource.activate_session(self.shop_api)

        if data.get('inventory_item_id'):
            spf.update_inventory_level(data.get('inventory_item_id'), data, 
                                       variant_id=product_id, **kwargs)
        elif product_id and kwargs.get('parent_id'):
            attr_list = kwargs.get('attr_list', ['size', 'color'])
            # print(data)
            if kwargs.get('formated') is None or kwargs['formated'] == False:
                model_to_spf.product_parser(data, 
                        attr_list,
                        kwargs.get('variation_attrs'),
                        parent_id=kwargs.get('parent_id'),
                        add_new=False
                        )
                spf.update_product(kwargs['parent_id'], product_id, data)
            else:
                spf.update_product(kwargs['parent_id'], product_id, data)
        
        if has_active_session == False:
            shopify.ShopifyResource.clear_session()


    def update_shop_product_batch(self, data:list, **kwargs):
        shopify.ShopifyResource.activate_session(self.shop_api)
        for product in data:
            self.update_shop_product(product, product['product_id'], has_active_session=True, 
                                                parent_id=product['parent_id'], **kwargs)
        shopify.ShopifyResource.clear_session()

    def sync_shop_orders(self,**kwargs):
        shopify.ShopifyResource.activate_session(self.shop_api)
        orders = []
        params = {'status': 'any'}
        if kwargs.get('latest_order_id'): params['since_id'] = kwargs['latest_order_id']
        if kwargs.get('status'): params['status'] = kwargs['status']
        if kwargs.get('fields'): params['fields'] = kwargs['fields']
        res =spf.get_all_orders(**params)
        for item in res:
            orders.append(spf_to_model.parse_order(item, self.field_mapping, **kwargs))

        shopify.ShopifyResource.clear_session()
        return orders

    def update_order_status_at_shop(self, destination_order_id: str, 
                                   status: str,**kwargs):
        has_active_session = kwargs.get('has_active_session', False)
        if has_active_session == False:
            shopify.ShopifyResource.activate_session(self.shop_api)
        
        if status not in ('cancel', 'cancelled'):
            logger.debug(f'status not support: {status}')
        else:
            spf.update_order_status(destination_order_id, 'cancel')

        if has_active_session == False:
            shopify.ShopifyResource.clear_session()
        

    def update_order_at_shop(self, destination_order_id: str, 
                             data, **kwargs):
        # has_active_session = kwargs.get('has_active_session', False)
        # if has_active_session == False:
        #     shopify.ShopifyResource.activate_session(self.shop_api)

        # spf.update_order()

        # if has_active_session == False:
        #     shopify.ShopifyResource.clear_session()
        pass

    def update_order_item_at_shop(self, destination_order_id: str, 
                             items: dict, **kwargs):
        """not supported in shopify (after order has been created)
           todo: graphAPI
        """
        pass
        # has_active_session = kwargs.get('has_active_session', False)
        # if has_active_session == False:
        #     shopify.ShopifyResource.activate_session(self.shop_api)
        
        # model_to_spf.transform_order_item(items)
        # from pprint import pprint
        # pprint(items)
        # spf.update_line_items(destination_order_id, items)

        # if has_active_session == False:
        #     shopify.ShopifyResource.clear_session()

    def get_order_at_shop(self, **kwargs):
        if kwargs.get('destination_order_id') is None:
            logger.debug('missing order id')
            return
        has_active_session = kwargs.get('has_active_session', False)
        if has_active_session == False:
            shopify.ShopifyResource.activate_session(self.shop_api)
        res = spf.get_order_info(kwargs.get('destination_order_id'))
        order = spf_to_model.parse_order(res, self.field_mapping, **kwargs)
        if has_active_session == False:
            shopify.ShopifyResource.clear_session()

        return order

        
    def get_products_at_shop(self,**kwargs):
        """page start from 0"""
        shopify.ShopifyResource.activate_session(self.shop_api)
        if self.product_ids is None:
            self.product_ids = spf.get_all_products(fields='id')
        logger.debug(f"total products {len(self.product_ids)}")
        batch = kwargs.get('batch_size', 5)
        start_i = kwargs.get('page', 0) * batch
        products = []
        for i in range(start_i, min(len(self.product_ids), start_i + batch)):
            products.append(self.get_product_at_shop(
                                 None, parent_id=self.product_ids[i]['id'], 
                                 has_active_session=True, **kwargs))
        shopify.ShopifyResource.clear_session()
        return products
                                    
    def get_product_at_shop(self, product_id: str, include_meta=True, **kwargs):
        """ get variant if parent_id is not passed, 
        otherwise pass product_id as None, it will get product"""
        has_active_session = kwargs.get('has_active_session', False)
        if has_active_session == False:
            shopify.ShopifyResource.activate_session(self.shop_api)

        product = None
        if kwargs.get('parent_id') is None:
            product = spf.find_variation(product_id, 
                                        include_meta=include_meta)
        else:
            product =spf.get_product_info(kwargs['parent_id'], 
                                        include_meta=include_meta)
        
        if has_active_session == False:
            shopify.ShopifyResource.clear_session()
       
        return spf_to_model.parse_product(product, self.field_mapping)
        
    
    def get_category_taxonomy(self):
        if self.collections is None: 
            self.get_all_collections()
        
        return [{'category': name} for name in self.collections.keys()]
    
    def update_customer_at_shop(self,destination_order_id: str, 
                             data: dict, **kwargs):
        shopify.ShopifyResource.activate_session(self.shop_api)
        model_to_spf.transform_customer(data)
        res, suc = spf.update_customer(destination_order_id, data, **kwargs)
        shopify.ShopifyResource.clear_session()
        if suc == True:
            return res 

        logger.debug(f'update customer failed: {res}')

    def add_tracking(self,destination_order_id: str, 
                            data: dict, **kwargs):
        """ff_info ={"location_id":66061795498,
            "tracking_numbers":["123456789"],
            "tracking_urls":["https://tracking.com/?abc"],
            "tracking_company": 'TBD',
            "notify_customer":False}"""
        shopify.ShopifyResource.activate_session(self.shop_api)
        ff_info = model_to_spf.transform_tracking(data, message=kwargs.get('message', True),
                                                add_new=kwargs.get('add_new', True))
        
        if kwargs.get('location_id') is None:
            #get location_id
            locations = spf.find_location()
            ff_info['location_id'] = locations[0]['id']
        else:
            ff_info['location_id'] = kwargs['location_id']

        _, do_update =spf.update_fulfillment_tracking(destination_order_id, 
                                                           ff_info,**kwargs)
        shopify.ShopifyResource.clear_session()

        return do_update

    def send_message(self, destination_order_id: str, 
                            message: str, **kwargs):
        """not supported by shopify"""
        pass


    def _parse_product(self, p, **kwargs):
        is_simple_product = kwargs.get('is_simple_product', False)
        if is_simple_product:
            data = p['variants'][0]
            data['destination_parent_id'] = str(p['id'])
            return data 
        elif p.get('variants') and len(p['variants']) > 0:
            data = {}
            for vari in p['variants']:
                vari['destination_parent_id'] = str(p['id'])
                data[vari['sku']] = vari 
            return data
        return p






import yaml

def db_yml_config(filename):
    """Return current db configuration in `yml`."""
    # todo validate db setting 
    with open(filename, 'r') as f:
        return yaml.safe_load(f)

def load_mapping(filename):
    with open(filename, 'r') as f:
        return yaml.safe_load(f)


def validate(db_setting):
    if db_setting.get('db_setting') is None:
        return False, 'can not find db_setting'

    if db_setting['db_setting'].get('db_type') is None:
        return False, 'can not find db_type'

    if db_setting['db_setting'].get('db_conn') is None:
        return False, 'can not find db_conn'

    conn = db_setting['db_setting']['db_conn']
    check_attrs = ['host', 'port', 'user', 'password', 'dbname']
    for attr in check_attrs:
        if conn.get(attr) is None:
            return False, f'can not find {attr} in db_conn'

    if db_setting['db_setting'].get('tables') is None:
        return False, 'can not find tables'

    tables = db_setting['db_setting']['tables']

    table_validation = {
        'category': ['category', 'full_path'],
        'product': ['product_name', 'identifier'],
        'front_shop': ['name', 'provider'],
        'variation': ['sku', 'title', 'upc', 'product_id'],
        'tax': ['name', 'country_code', 'default_rate'],
        'destination': ['destination_product_id', 'price', 'currency',
                    'is_current_price', 'sku', 'front_shop_id', 'available'],
        'inventory': ['sku', 'cost', 'currency', 'supplier_item_id', 'supplier_id'],
        'customer': ['first_name','last_name','phone', 'email', 'address1',
                    'address2', 'country', 'state', 'city', 'postcode', 
                    'is_shipping', 'is_billing'],
        'shop_order': ['destination_order_id', 'front_shop_id', 'status',
                     'billing_customer_id','shipping_customer_id',
                     'payment_type','payment_token', 'refund', 'shipping_cost',
                     'shipping_tax_id', 'order_total','order_at','currency'],
        'order_item': ['shop_order_id', 'destination_id','destination_product_id',
                      'tax_id', 'order_qty', 'ship_qty'],
        'fulfillment': ['provider', 'status', 'dimension', 'weight', 
                      'tracking_id', 'shop_order_id', 'destination_order_id']
    }

    for k, val in table_validation.items():
        if tables.get(k) is None:
            return False, f'missing table {k}'
        
        if tables[k].get('columns') is None:
            return False, f'{k} missing columns'

        cols = [c['name'] for c in tables[k]['columns']]
        for col in val:
            if col not in cols:
                return False, f'{k} missing column {col}'

    return True, None


            


import click 
from .cli import cli
from pypipet.core.operations.utility import get_front_shop_id
from pypipet.core.operations.frontshop import update_product_to_db
from pypipet.core.operations.frontshop import update_variation_to_db
from pypipet.core.operations.frontshop import update_destination_to_db
from pypipet.core.operations.frontshop import update_product_at_front_shop, update_product_at_front_shop_bulk
from pypipet.core.operations.frontshop import  add_product_to_shop
from pypipet.core.operations.frontshop  import get_product_with_variations,get_product_info
from pypipet.core.sql.query_interface import  search_exist
from pypipet.core.operations.inventory import get_inventory_by_sku, update_instock_front_shop
from pypipet.core.fileIO.file_loader import read_csv_file
from .utility import col2dict_update,map_product_sku_for_launch
from pprint import pprint

@cli.command()
@click.argument("action", type=click.Choice(['launch', 'deactivate', 'activate',
                                             'edit', 'show']))
@click.option("--shop", 
                help="frontshop name (in setting.shops)")
@click.option("--sku",
                help="if multiple sku provided, seperate with comma")
@click.option("-f", "--filename",
                help=" edit product attributes")
@click.option("--price",
                help=" price for lauching product to frontshop")
@click.option("--currency", default='USD',
                help=" currency for lauching product to frontshop")
@click.option("--include-stock", is_flag=True,
                help=" update stock for lauching product to frontshop(shopify only)")
@click.option('--skip-shop', is_flag=True, 
                        help="skip updating front shop for action edit.")
@click.pass_context
def product(ctx, action, shop, sku, price, currency, filename, include_stock,skip_shop):
    """
    add/edit products (by sku or file)
    """
    click.echo('{} products at {}. skip-shop {}'.format(
        action, shop, skip_shop))
    if sku is None and filename is None:
        click.echo('product identifier not found')
        return

    if sku is not None:
        sku = sku.split(',')

    project = ctx.obj['project']
    config = ctx.obj["config"]
    if config is None:
        click.secho('create project first',
                   fg='yellow')
        return

    project.initialize_project(config)
    table_classes = project.get_table_objects()
    sessionmaker = project.get_session_maker()
    session = sessionmaker()
    shop_conn = project.get_shop_connector(shop)

    if shop_conn is None:
        click.secho('shop setting not found: {}'.format(shop),
                   fg='yellow')
        session.close()
        return 
    get_front_shop_id(table_classes, session, shop_conn)

    if action == 'launch':
        #to do, launch by file
        if sku is None and filename is None:
            click.echo('missing feeds')
            return 
        if sku and price is None:
            click.echo('missing price')
            return
        
        if filename is not None:
            click.echo('launch by file {}'.format(filename))
            _launch_to_shop_by_file(session, table_classes, shop_conn, 
                                currency,
                                project.config.get('attr_list'),
                                project.config.get('variation_attrs'),
                                filename,
                                schema=project.schema) 
        else: 
            price = float(price)
            _launch_to_shop(session, table_classes, shop_conn, 
                                    sku, price, currency,
                                    project.config.get('attr_list'),
                                    project.config.get('variation_attrs'))
        if include_stock:
            click.echo('updating instock qty...')
            _update_stock_for_launching(session, table_classes, shop_conn)
            
    elif action == 'deactivate':
        if sku is None:
            click.echo('missing sku')
            return
        _change_status(session, table_classes, shop_conn, 
                                       sku, False, skip_shop)
    elif action == 'activate':
        if sku is None:
            click.echo('missing sku')
            return
        _change_status(session, table_classes, shop_conn, 
                                       sku, True, skip_shop)
    elif action == 'edit':
        if filename is None:
            click.echo('missing data file')
            return
        prods = _get_data_from_file(filename, action)
        click.echo('total product to edit {}'.format(len(prods)))
        _edit_product(session, table_classes, 
                      shop_conn, prods, skip_shop)
    elif action == 'show':
        if sku is None:
            click.echo('missing sku')
            return
        for s in sku:
            variation_info = get_product_with_variations(
                    table_classes, session, 
                    {'sku': s}, 
                     include_published=True, 
                     front_shop_id=shop_conn.front_shop_id,
                     schema=project.schema)
            click.echo(variation_info)
    else:
        click.secho('action not supported: {}'.format(action),
                fg='yellow')
    
    session.close()

def _launch_to_shop_by_file(session, table_classes, shop_conn,  
                            currency, attrs, variation_attrs, filename, schema=''):
    data = _get_data_from_file(filename, 'launch')
    
    for sku_price in data:
        product_info = get_product_info(table_classes, session, shop_conn.front_shop_id, 
                    sku=sku_price['sku'],
                    include_category=True,
                    schema=schema)
        res = add_product_to_shop(table_classes, session, shop_conn, product_info, 
                                        price=sku_price['price'],
                                        currency=currency,
                                        attr_list = attrs,
                                        variation_attrs=variation_attrs
                                        )
                             
        click.echo('sku {} add to shop.. success.. {}'.format(sku_price['sku'], res is not None))


def _launch_to_shop(session, table_classes, shop_conn, skus, 
                            price, currency, attrs, variation_attrs, schema=''):
    for sku in skus:
        product_info = get_product_info(table_classes, session, shop_conn.front_shop_id, 
                    sku=sku,
                    include_category=True,
                    schema=schema)
        res = add_product_to_shop(table_classes, session, shop_conn, product_info, 
                                        price=price,
                                        currency=currency,
                                        attr_list = attrs,
                                        variation_attrs=variation_attrs
                                        )
                             
        click.echo('sku {} add to shop.. success.. {}'.format(sku, res is not None))

def _update_stock_for_launching(session, table_classes, shop_conn):
    update_instock_front_shop(table_classes, session, shop_conn)

def _change_status(session, table_classes, shop_conn, 
                        skus, status, skip_shop):
    for sku in skus:
        update_destination_to_db(table_classes.get('destination'),
                                     session,
                                     {'available': status, 'sku': sku},
                                     shop_conn.front_shop_id
                                     )
        
        if skip_shop is False:
            update_data = {'destination': {'available': status}}
            inv = 0
            if status:
                qty = get_inventory_by_sku(table_classes,
                                           session, sku)
                if qty: inv = qty

            update_data['variation'] ={
                'in_stock': inv
            }
            res = update_product_at_front_shop(
                                 table_classes, 
                                 session, 
                                 shop_conn, 
                                 update_data, 
                                 sku) 
            if res is None:
                click.echo('sku {} update error'.format(sku))

def _edit_product(session, table_classes, shop_conn, 
                                       products, skip_shop):
    for p in products:
        click.echo('updating sku {}'.format(p['variation']['sku']))
        if len(p['product']) > 0:
            if p['product'].get('id') is None:
                product_id = _get_product_id_by_sku(table_classes.get('variation'), 
                                                   session, p['variation']['sku'])
                if product_id is None:
                    click.echo('invalid sku, product not found')
                    return
                p['product']['id'] = product_id

            update_product_to_db(table_classes, 
                           session, 
                           p['product'])
        
        if len(p['variation']) > 0:
            update_variation_to_db(table_classes.get('variation'), 
                           session, 
                           p['variation'])
        
        if len(p['destination']) > 0:
            if p['destination'].get('sku') is None:
               click.secho('missing destination sku: {}'.format(p),
                fg='yellow')  
               return 
            update_destination_to_db(table_classes.get('destination'),
                                     session,
                                     p['destination'],
                                     shop_conn.front_shop_id
                                     )
        
    if skip_shop is False:
        click.echo('updating frontshop, total {}'.format(len(products)))
        update_product_at_front_shop_bulk(table_classes, session, 
                                shop_conn, products, batch_size=50)
        

def _get_data_from_file(filename, action):
    data = read_csv_file(filename)
    if action == 'launch':
        return map_product_sku_for_launch(data) 
    else:
        products = []
        #validation
        for i, row in data.iterrows():
            sku = str(row['variation.sku'])
            if sku is None or sku.strip() == '':
                click.secho('missing {}, stop processing'\
                                    .format(sku), 
                            fg='yellow')
                return 

        for i, row in data.iterrows():
            sku = str(row['variation.sku'])
            p = col2dict_update(sku, row)
            # pprint(p)
            products.append(p)

        return products


def _get_product_id_by_sku(table_obj, session, sku):
    vari = search_exist(table_obj, session, {'sku': sku})
    if len(vari) > 0:
        return vari[0].product_id



import click 
from .cli import cli


@cli.command()
@click.argument("project_name")
@click.pass_context
def new(ctx,project_name):
    """
    create new project
    """
    click.echo('initialize project: {}'.format(project_name))
    res = ctx.obj['project'].start_project(project_name)
    if not res:
        click.secho('project setting existed in curent dir',
                   fg='yellow',)

@cli.command()
@click.pass_context
def initialize(ctx):
    """
    initialize project
    """
    click.echo('initialize project')
    click.echo("home dir: {}".format(ctx.obj['project'].root))
    
    config = ctx.obj["config"]
    
    # print(config)
    ctx.obj['project'].initialize_project(config)
    ctx.obj['project'].import_static_data()





from pypipet.core.sql.query_interface import get_server_timestamp
from pypipet.core.sql.query_interface import add_json_to_db, search_exist
from pypipet.core.sql.query_interface import update_data, update_inventory,update_bulk
from pypipet.core.sql.query_interface import get_variation_instock_qty
from pypipet.core.sql.query_interface import aggregate_inventory_qty
from pypipet.core.model.product import Inventory
from .utility import _object2dict
import logging
from datetime import datetime, timedelta

logger = logging.getLogger('__default__')


def get_supplier_id(table_obj, session, params):
    suppliers = search_exist(table_obj, 
                             session, 
                             params)
    if len(suppliers) >0:
        return suppliers[0].id 



def update_inventory_bulk(table_objs, session, invs:list, 
                                   batch_size=500, ignore_new=True):
    """ignore_new: if sku already in inventory table """
    if ignore_new:
        #require id to update in bulk query
        if invs and len(invs) > 0:
            if invs[0].get('id') is None:
                logger.debug(f"missing id in data to update in bulk")
                return 
        update_bulk(table_objs.get('inventory'), session, invs)
        update_instock_qty_db(table_objs, session, batch_size=batch_size)
        return invs
    
    sku_qty = {}
    for inv in invs:
        params ={
            'supplier_id': inv['supplier_id'],
        }
        if inv.get('sku') is not None:
            params['sku'] = inv['sku']
        if inv.get('supplier_item_id') is not None:
            params['supplier_item_id'] = inv['supplier_item_id']
        else:
            params['supplier_item_id'] = inv['sku']
        
        logger.debug(f"processing inv {inv}")
        update_inventory(table_objs.get('inventory'), session, inv, params)

        if sku_qty.get(inv['sku']) is None:
            sku_qty[inv['sku']] = {'sku': inv['sku'], 'in_stock': inv['qty']}
        else:
            sku_qty[inv['sku']]['in_stock'] += inv['qty']

    logger.debug(f"updating variation table")
    #update variation table
    update_bulk(table_objs.get('variation'), session, list(sku_qty.values()))
    return invs


def update_inventory_db_by_sku(table_objs, session, sku, inv:dict, params:dict =None):
    """update inventory table with supplier info"""
    if params is None:
        params={'sku': sku}
    if params.get('supplier_item_id') is None:
        params['supplier_item_id'] = sku
    if inv.get('supplier_id') is not None:
        params['supplier_id'] = inv['supplier_id']

    if inv.get('qty') is None:
        inv['qty'] = 0
    
    update_inventory(table_objs.get('inventory'), session, inv, params)
    sku_inv = get_inventory_by_sku(table_objs, session, sku, by_supplier=True)
    if sku_inv is None:
        logger.debug(f'inventory not found sku {sku}')
        return
    qty = sum([item['qty'] for item in sku_inv])
    return update_data(table_objs.get('variation'), 
                        session, 
                        {'in_stock': qty}, 
                        {'sku': sku})

def update_inventory_order_processing(table_objs, session, sku, sale_qty, supplier_id=None):
    """after process order, update the inventory
        if not managed with supplier, update instock(variation) only 
    """
    sku_inv = get_inventory_by_sku(table_objs, session, sku, by_supplier=True)

    if supplier_id:
        supplier_inv = list(filter(lambda inv: (inv['supplier_id']==supplier_id), sku_inv))
        if len(supplier_inv) > 0:
            update_inventory(table_objs.get('inventory'), session, {'qty': supplier_inv[0]['qty'] - sale_qty}, 
                                        {'id': supplier_inv[0]['id']})
        
    in_stock = sum([item['qty'] for item in sku_inv]) - sale_qty
    update_data(table_objs.get('variation'), 
                    session, 
                    {'in_stock': in_stock}, 
                    {'sku': sku})
    return in_stock

def update_instock_qty_db(table_objs, session, batch_size=50, 
                        params:dict=None, latest_hours=24):
    """aggregate instock from inventory table"""
    if params is None: params = {}
    start_sku = None
    while True:
        logger.debug(f'start from {start_sku}')
        batch = aggregate_inventory_qty(table_objs, session, 
                       start_sku, params=params, 
                       batch_size=batch_size, latest_hours=latest_hours)

        if batch is None or len(batch) == 0:
            logger.debug('done')
            break
       
        update_bulk(table_objs.get('variation'), session, batch)
        # for rd in batch:
        #     update_data(table_objs.get('variation'), 
        #                 session, 
        #                 {'in_stock': rd['qty']}, 
        #                 {'sku': rd['sku']})

        start_sku = batch[-1]['sku']



def update_instock_front_shop(table_objs, session, shop_connector,
                        batch_size=50, params:dict=None,latest_hours=None, **kwargs):
    """
    update all product at front shop instock qty
    """
    if params is None: params = {'front_shop_id': shop_connector.front_shop_id}
    pid = None
    while True:
        res = get_variation_instock_qty(table_objs, session, 
                    pid, params=params, batch_size=batch_size, 
                    latest_hours=latest_hours)
        if res is None or len(res) == 0:
            logger.debug('done')
            break
        
        logger.debug(f"updating {res[0]['product_id']}, batch {len(res)}")
        pid = res[-1]['product_id']
        shop_connector.update_shop_product_batch(res, **kwargs)
        # print(f'last update {pid} ')

def update_instock_front_shop_by_sku(table_objs, session, shop_connector,
                                sku, params:dict=None, **kwargs):
    if params is None: params = {}
    params.update({'sku': sku, 
                   'front_shop_id': shop_connector.front_shop_id})
    dest = search_exist(table_objs.get('destination'), session, params)
    if dest and len(dest) >0:
        dest = dest[0]
        qty = get_inventory_by_sku(table_objs, session, sku)
        if qty is None:
            logger.debug(f"invalid sku{sku}")
            return 
        shop_connector.update_shop_product(
                        {
                        'qty': qty,
                        'sku': sku},
                        dest.destination_product_id,
                        parent_id=dest.destination_parent_id,
                        **kwargs)
        logger.debug(f"updated product {dest.destination_product_id}")
    else:
        logger.debug(f"invalid sku {sku}")
        
def get_inventory_by_sku(table_objs, session, sku, by_supplier=False):
    if not by_supplier:
        inv = search_exist(table_objs.get('variation'), 
                                session, 
                                {'sku': sku})
        if len(inv) >0:
            return inv[0].in_stock 
    else:
        inv = search_exist(table_objs.get('inventory'), 
                                session, 
                                {'sku': sku})
        if len(inv) >0:
            return [_object2dict(r) for r in inv]

def match_variation_sku_by_upc(table_obj, session, invs:list):
    for i, inv in enumerate(invs):
        if inv['upc'] is not None and inv.get('sku') is None:
            exist = search_exist(table_obj, session, {'upc': inv['upc']})
            if exist and len(exist) > 0:
                invs[i]['sku'] = exist[0].sku
    return invs
    
# def update_instock_qty_db_simple(table_objs, session, inv_list: [dict]):
#     """if do not inventory table, update by sku"""
#     for rd in inv_list:
#         if rd.get('sku') is not None and rd.get('qty') is not None:
#             update_data(table_objs.get('variation'), 
#                         session, 
#                         {'in_stock': rd['qty']}, 
#                         {'sku': rd['sku']})

def update_instock_to_variation_db(table_obj, session, sku, qty):
    instock = search_exist(table_obj, 
                        session, 
                        {'sku': sku})
    if instock is None or len(instock) == 0:
        logger.debug(f'sku does not exist {sku}')
        return 

    updated_qty = max(0, instock[0].in_stock - qty)
    assert isinstance(qty, int)
    update_data(table_obj, 
                session, 
                {'in_stock': updated_qty}, 
                {'sku': sku})
    return updated_qty


from pypipet.core.sql.query_interface import search_exist, update_data
from pypipet.core.sql.query_interface import get_fulfilled_orders
from pypipet.core.sql.query_interface import add_json_to_db
import logging
from .utility import _object2dict
# from datetime import datetime, timedelta

logger = logging.getLogger('__default__')
logger.setLevel(logging.DEBUG)


def get_shipping_info(table_objs, session, params:dict):
    shipment = {}
    ff_params = {}
    if (params.get('destination_order_id') is None\
        or params.get('front_shop_id') is None) \
        and params.get('shop_order_id') is None:

        if params.get('tracking_id') is None:
            logger.debug('invalid params')
            return

        ff_params = {'tracking_id': params['tracking_id']}
    elif params.get('shop_order_id') is not None:
        ff_params = {'shop_order_id': params['shop_order_id']}

    if len(ff_params) > 0:
        ff = get_fulfillment(table_objs.get('fulfillment'), session, ff_params)
        shipment.update(ff)
        if shipment.get('status') is not None:
            shipment['shipping_status'] = shipment['status'] 
            params['shop_order_id'] = shipment['shop_order_id']
    
    if params.get('shop_order_id') is not None:
        params = {'id': params['shop_order_id']}

    order = search_exist(table_objs.get('shop_order'), 
                session, 
                params)
    
    if len(order) == 0:
        logger.debug('order not exist')
        return None 

    if len(ff_params) == 0:
        ff =  get_fulfillment(table_objs.get('fulfillment'), session, 
                             {'shop_order_id': order[0].id})
        shipment.update(ff)
        if shipment.get('status') is not None:
            shipment['shipping_status'] = shipment['status'] 
    
    shipment.update(_object2dict(order[0]))

    #shipping address 
    customer = search_exist(table_objs.get('customer'), 
                   session, 
                   {'id': shipment['shipping_customer_id']})
    if len(customer) > 0:
        customer = _object2dict(customer[0])
        del customer['id']
        shipment.update(customer)

    return shipment


def get_unfulfilled_order_inprocesss(table_objs, session, status='processing', 
                            params:dict=None, schema=''):
    if params is None: params = {}
    params.update({'status': status})
    in_process = search_exist(table_objs.get('shop_order'), session,params )
    if len(in_process) > 0:
        in_process_orders = [_object2dict(r) for r in in_process]
        shop_order_id_list = [r['id'] for r in in_process_orders]
        fulfilled = get_fulfilled_orders(table_objs, session, status=status,
                shop_order_id_list=shop_order_id_list, schema=schema)
        
        fulfilled_order_ids =set( [r['shop_order_id'] for r in fulfilled])
       
        unfilled_orders = []
        for order in in_process_orders:
            if not order['id'] in fulfilled_order_ids:
                unfilled_orders.append(order)

        return unfilled_orders
    logger.debug(f'invalid params {params}')

def get_fulfilled_orders_inprocess(table_objs, session, status='processing', 
                           params:dict=None, schema=''):
    if params is None: params = {}
    params.update({'status': status})
    fulfilled = get_fulfilled_orders(table_objs, session, status=status,
                params=params, schema=schema)
        

    return fulfilled


def add_fulfillment_bulk(table_objs, session, data_list:list):
    for data in data_list:
        add_fulfillment(table_objs, session, data)

def add_fulfillment(table_objs, session, data:dict, **kwargs):
    if data.get('tracking_id') is  None \
        or data.get('provider') is  None :
        logger.debug('invalid fulfiment info')
        return 
    if (data.get('destination_order_id') is None\
        or data.get('front_shop_id') is None) \
        and data.get('shop_order_id') is None:
        logger.debug('invalid identifier info')
        return     
    if data.get('shop_order_id') is None:
        order = _get_order(table_objs.get('shop_order'), 
        session, 
        {
            'destination_order_id': data['destination_order_id'],
            'front_shop_id': data['front_shop_id']
        })
        if order:
            data['shop_order_id'] = order.id
            del data['front_shop_id']

    if data.get('status') is None: data['status'] = 'shipped'

    existing_ff = search_exist(table_objs.get('fulfillment'),session, 
                               {'shop_order_id': data['shop_order_id']})
    
    if existing_ff and len(existing_ff) > 0:
        if kwargs.get('update_exist') == True:
            logger.debug('updating existing tracking')
            return update_data(table_objs.get('fulfillment'), session, 
                            data, {'shop_order_id': data['shop_order_id']})
        else:
            logger.debug(f"existing tracking info . update_exist flag\
                {kwargs.get('update_exist', False)}")
    else:
        return add_json_to_db(table_objs.get('fulfillment'), session, data)

def update_tracking(table_objs, session, data:dict, params: dict):
    if (params.get('destination_order_id') is None\
        or params.get('front_shop_id') is None) \
        and params.get('shop_order_id') is None:
        logger.debug('invalid identifier info')
        return 
    
    if params.get('shop_order_id') is None:
        order = _get_order(table_objs.get('shop_order'), 
        session, 
        {
            'destination_order_id': params['destination_order_id'],
            'front_shop_id': params['front_shop_id']
        })
        if order:
            params['shop_order_id'] = order.id
            del params['front_shop_id']

    if data.get('status') is None: data['status'] = 'shipped'
    update_data(table_objs.get('fulfillment'), session, data, params)

def get_fulfillment(table_obj, session, params):
    ff = search_exist(table_obj,session, params)
    if len(ff) > 0:
        return  _object2dict(ff[0])
    
    # logger.debug('existing fulfillment not found')
    return {}


def update_fulfillment_to_front_shop(table_objs, session, 
             shop_connector, params: dict, tracking_info:dict=None, **kwargs):
    if tracking_info is None or len(tracking_info) == 0:
        if (params.get('destination_order_id') is None \
        or params.get('front_shop_id') is None) and params.get('shop_order_id') is None:
            logger.debug('invalid identifier info')
            return

        if params.get('shop_order_id'):
            params['id'] = params['shop_order_id']
            del params['shop_order_id']

        order = _get_order(table_objs.get('shop_order'),session, params)
        if order:
            tracking_info = get_fulfillment(table_objs.get('fulfillment'), 
                                session,
                                {'shop_order_id': order.id})
        else:
            return
        
    res = shop_connector.add_tracking(
                tracking_info['destination_order_id'],
                tracking_info,
                **kwargs)
    return res



def _get_order(table_obj, session, params):
    order = search_exist(table_obj, 
                session, 
                params)
    
    if order is None or len(order) == 0:
        logger.debug('order not exist')
        return None 
    
    return order[0]


from pypipet.core.sql.query_interface import *
from pypipet.core.model.product import Destination, Product
from .utility import get_front_shop_id
from copy import deepcopy

import logging
logger = logging.getLogger('__default__')

_DEBUG_ = True


def add_taxonomy_to_db(table_obj, session, shop_connector):
    categories = shop_connector.get_category_taxonomy()
    if categories:
        for cat in categories:
            db_insert_raw(table_obj, session, cat)
    else:
        logger.info('fetching taxonomy failed')


def load_products_from_shop(table_objs, session, shop_connector, 
                         start_from=1,  **kwargs):
    start_page = start_from
    while True:
        logger.debug(f'page {start_page}')
        res = shop_connector.get_products_at_shop(
                                  batch_size=shop_connector.batch_size, 
                                  page=start_page,
                                  **kwargs
                                  )
        if _DEBUG_ : logger.debug(f"total {len(res)}")
        if res is None or len(res) == 0:
            logger.debug('done sync')
            break
        
        add_product_to_db_bulk(table_objs, session, 
                                shop_connector.front_shop_id,res)
        start_page += 1
        if _DEBUG_ : break


def get_product_from_shop(shop_connector, 
                                destination_product_id, **kwargs):
    res = shop_connector.get_product_at_shop(destination_product_id,
                                             **kwargs)
    return res

def add_product_to_db_bulk(table_objs, session, 
                    front_shop_id,products:list, currency='USD',**kwargs):
    for p in products:
        # if _DEBUG_: logger.debug(p)
        if p is None: continue
        add_product_with_variations_to_db(table_objs, session, 
                            front_shop_id, p, currency=currency, **kwargs)

def add_product_with_variations_to_db(table_objs, session, 
                            front_shop_id,p:dict, currency='USD',**kwargs):
    add_product_to_db(table_objs, session, p['product'], **kwargs)
    if p.get('destinations') is None \
                or len(p['destinations']) == 0: 
        return
    for dest in p['destinations']:
        #print(p['destinations'])
        add_destination_to_db(table_objs, 
                            session, 
                            front_shop_id, 
                            dest,
                            currency=currency)


def add_product_to_db(table_objs, session, product_info: dict,**kwargs):
    if product_info is None: return 
    product = Product()
    #check category
    if product_info.get('category') is not None \
    and product_info.get('category').strip() != '' \
    and type(product_info.get('category')) is str:
        # print(product_info['product_name'])
        cat = _get_category_by_name(
                            table_objs.get('category'), 
                            session, 
                            product_info['category'])
        
        product_info.update(cat)
        del product_info['category']
    
    product.set_product(table_objs, product_info, **kwargs)
    
    return add_new_product(table_objs, session, product)




def add_destination_to_db(table_objs, session, front_shop_id, 
                                        dest_data, **kwargs):
    
    dest_data.update({
        'front_shop_id': front_shop_id,
        'is_current_price': True,
        'currency': kwargs.get('currency', 'USD')
    })
    if dest_data.get('available') is None:
        dest_data['available'] = True 
    
    ignore_exist = kwargs.get('ignore_exist', False)
    if not ignore_exist:
        exist = search_exist(table_objs.get('destination'), 
                                    session, {
                                        'sku': dest_data['sku'],
                                        'front_shop_id': front_shop_id
                                    })
        if exist and len(exist) > 0:
            #update 
            # update_data(table_objs.get('destination'),  session,
            #         dest_data, {'id': exist[0].id})
            logger.debug(f'destination exists {dest_data}')
            return 
   
    #add new 
    dest = Destination()
    dest.set_destination(table_objs, dest_data)
    
    add_to_db(table_objs.get(dest.__table_name__), session, dest)

def is_exist_destination(table_obj, session, dest_data):
    dest = search_exist(table_obj, session,{
        'front_shop_id': dest_data['front_shop_id'],
        'sku': dest_data['sku']
    })
    if dest is None or len(dest) == 0:
        return False 
    else:
        for d in dest:
            if d.is_current_price:
                logger.debug(f'data exist id {d.id}')
                dest_data['id'] = d.id
                return True 
        return False



def get_product_info(table_objs, session, front_shop_id, identifier=None,
                    sku=None, product_id=None, include_published=False,
                    include_category=False, **kwargs):
    params = None
    if identifier is not None:
        params = {'identifier': identifier}
    elif product_id is not None:
        params = {'product_id': product_id}
    elif sku is not None:
        params = {'sku': sku}
    if params is None:
        logger.debug('invalid product params')
        return 
    
    variation_info = get_product_with_variations(
                     table_objs, 
                     session, 
                     params, 
                     include_published=include_published,
                     include_category=include_category, 
                     front_shop_id=front_shop_id,
                     schema=kwargs.get('schema', ''))
    return variation_info

def add_product_to_shop(table_objs, session, shop_connector, product_info, 
                                    price=None, prices:dict=None, **kwargs):
    if product_info is None: return 
    
    if product_info.get('destinations'):
        logger.debug(f"has destinations, {product_info['destinations']}")
        return
    if (price is None and prices is None) or \
    (prices is not None and type(prices) is not dict):
        logger.debug('invalid product params')
        return 
    
    currency = kwargs.get('currency', 'USD')
    if product_info.get('variations') is None:
        #product without variations 
        product_info['price'] = price
        product_info['currency'] = currency
        res = shop_connector.add_product_to_shop(
                                             deepcopy(product_info),
                                             **kwargs) 

        if res is None:
            logger.debug(f"adding product to front shop failed {product_info['sku']}")
            return None
        #add to database
        res.update({
            'sku': product_info['sku'],
            'price': price,
            'destination_product_id': str(res['id'])
        })
        add_destination_to_db(table_objs, session, 
                             shop_connector.front_shop_id, 
                             res, currency=currency,
                             ignore_exist=kwargs.get('ignore_exist', True))
        return res
    else:
        for i, vari in enumerate(product_info['variations']):
            product_info['variations'][i]['currency'] = currency 
            if price: 
                product_info['variations'][i]['price'] = price 
            else:
                if prices.get(vari['sku']) is None:
                    logger.debug(f"missing price for sku {vari['sku']}")
                product_info['variations'][i]['price'] = \
                             prices.get(vari['sku'])
       
        assert kwargs.get('variation_attrs') is not None
        
        parent_id = kwargs.get('parent_id', product_info.get('parent_id'))
        
        res = shop_connector.add_product_to_shop(
                                             deepcopy(product_info),
                                             parent_id=parent_id,
                                             **kwargs) # return dict to res
        if res is None or len(res) == 0:
            logger.debug(f"adding product to front shop failed {product_info['identifier']}")
            return None
        for i, vari in enumerate(product_info['variations']):
            if res.get(vari['sku']):
                if price is None:
                    price =  prices.get(vari['sku'])

                dest_data= res[vari['sku']]
                dest_data.update({
                    'sku': vari['sku'],
                    'price': price,
                    'destination_product_id': str(dest_data['id'])
                })
                add_destination_to_db(table_objs, session, 
                             shop_connector.front_shop_id, 
                             dest_data, currency=currency,
                             ignore_exist=kwargs.get('ignore_exist', True))
                # dest_data.update({
                #     'sku': vari['sku'],
                #     'front_shop_id': shop_connector.front_shop_id,
                #     'price': price,
                #     'is_current_price': True,
                #     'available': True,
                #     'currency': currency,
                # })
                # exists = search_exist(table_objs.get('destination'), session, 
                #                      {'sku': vari['sku'],
                #                       'front_shop_id': shop_connector.front_shop_id,
                #                       'is_current_price': True,
                #                       'available': True})
                # if exists and len(exists) > 0:
                #     update_data(table_objs.get('destination'), session, 
                #                 dest_data,
                #                 {'id': exists[0].id})
                # else:
                #     add_json_to_db(table_objs.get('destination'), session, 
                #                    dest_data)
            else:
                logger.debug(f"missing results for sku {vari['sku']}")  
        return res  

def add_variation_to_shop(table_objs, session, shop_connector, 
                             sku:str, price:float, **kwargs):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    
    #transform data
    # variation_info = get_variation_info_by_sku(table_objs, session, sku, 
    #                  front_shop_id=shop_connector.front_shop_id)
    product_info = get_product_with_variations(table_objs, session, 
                    {'sku': sku}, 
                     include_published=False, 
                     front_shop_id=shop_connector.front_shop_id,
                     schema=kwargs.get('schema', ''))
    
    if product_info is None:
        return 
    
    #check if it has been added but available_flag need update to True
    if product_info.get('set_available_flag'):
        #activate again 
        update_data(table_objs.get('destination'), 
                    session, 
                    {'available': True}, 
                    {'id': product_info['destination_id'] })
        logger.info(f"variation sku {sku} reactivated")
            
        return 

    if product_info.get('destinations') is not None:
        del product_info['destinations']
    
    # if the sku is one of the variations of a product
    if product_info.get('variations') and type(product_info['variations']) is list:
        if kwargs.get('ignore_variations'):
            product_info.update(product_info['variations'][0])
            del product_info['variations']
        else:
            logger.info('add as variation')
            return add_product_to_shop(table_objs, session, shop_connector, 
                       product_info, price=price, **kwargs)

    #add to frontshop 
    product_info['price'] = price
    product_info['currency'] = kwargs.get('currency', 'USD')
    res = shop_connector.add_product_to_shop(
                                             product_info,
                                             **kwargs)
    # if res is None:
    #     #if exist, update
    #     res = _update_product_at_frontshop_by_sku(table_objs, session, 
    #                                       shop_connector, 
    #                                       sku, product_info)
    # 
       
    if res is None:
        logger.debug(f'adding product to front shop failed {sku}')
        return None
    #add to database
    dest_data= res
    dest_data.update({
        'sku': sku,
        'price': price,
        'destination_product_id': str(dest_data['id'])
    })
    add_destination_to_db(table_objs, session, 
                    shop_connector.front_shop_id, 
                    dest_data, currency=product_info['currency'],
                    ignore_exist=kwargs.get('ignore_exist', True))
    
    return res


def get_cost_by_sku_from_db(table_obj, session, sku, params: dict=None):
    """apply when  skus have multiple suppliers"""
    if params is None: params = {}
    params['sku'] = sku
    exists = search_exist(table_obj, session, params)
    if len(exists) > 0:
        prices = [inv.cost for inv in exists]
        return sum(prices)/len(prices)

def update_product_to_db(table_objs, session, product_info: dict):
    params = {}
    if product_info.get('id') is not None:
        params['id'] = product_info['id']
    elif product_info.get('identifier') is not None:
        params['identifier'] = product_info['identifier']
    else:
        logger.debug('missing product id and identifier')
        return 
    #check category
    if product_info.get('category') is not None \
    and product_info.get('category').strip() != '' \
    and type(product_info.get('category')) is str:
        # print(product_info['product_name'])
        cat = _get_category_by_name(
                            table_objs.get('category'), 
                            session, 
                            product_info['category'])
        
        product_info.update(cat)
        del product_info['category'] 
    
    if product_info.get('short_description'):
        product_info['short_description'] = product_info['short_description'][:1000]
    update_data(table_objs.get('product'), session, product_info, 
                params)    

def update_variation_to_db(table_obj, session, variation_info: dict):
    update_data(table_obj, session, variation_info, 
                {'sku': variation_info['sku']})     

def update_destination_to_db(table_obj, session, dest_info: dict, front_shop_id):
    params ={
        'sku': dest_info['sku'],
        'front_shop_id': front_shop_id,
        'is_current_price': True
    }
    if dest_info.get('price') is not None:
        destination = update_destination_price(table_obj, 
                                    session, dest_info['price'], params)
    
    update_data(table_obj, session, dest_info, params)  

def update_front_shop_price_bulk(table_objs, session, shop_connector, data: list):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    for d in data:
        if d.get('price') is None: continue
        price = d['price']
        del d['price']

        d['front_shop_id'] = shop_connector.front_shop_id
        update_front_shop_price(table_objs, session, 
                            shop_connector, price, d)
    return data

def update_front_shop_price(table_objs, session, shop_connector, price:float, params:dict):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    #udpate database
    destination = update_destination_price(table_objs.get('destination'), 
                                              session, price, params)
    # print(destination.destination_product_id)
    if destination:
        #udpate front shop
        res = shop_connector.update_shop_product(
                            {'price': price}, 
                            product_id=destination.destination_product_id,
                            parent_id=destination.destination_parent_id
                            )
        
        return res

def update_product_at_front_shop_bulk(table_objs, session, shop_connector, 
                                   data_list: list, batch_size=50,  **kwargs):
    """bulk option not available for shopify"""
    
    update_batch = []
    

    for data in data_list:
        product_info = {}
        product_id = None
        parent_id = None
        sku = data['variation']['sku']
        if data.get('destination') is None or \
        data['destination'].get('destination_product_id') is None:
            exists = search_exist(table_objs.get('destination'), 
                                session, 
                                {'sku': sku,
                                'front_shop_id': shop_connector.front_shop_id,
                                #    'available': True,
                                'is_current_price': True})
            if exists is None or len(exists) == 0:
                logger.info(f"variation sku {sku} not at shop or not available")
                continue
            product_id = exists[0].destination_product_id
            parent_id = exists[0].destination_parent_id
            if data['destination'].get('price') is None:
                data['destination']['price'] = exists[0].price
        else:
            product_id = data['destination']['destination_product_id']
            parent_id = data['destination'].get('destination_parent_id')
        
        product_info.update(data.get('product', {}))
        product_info.update(data.get('variation', {}))
        product_info.update(data.get('destination', {}))
        product_info['id'] = product_id
        product_info['parent_id'] = parent_id
        # print(product_info)
        update_batch.append(product_info)

        if len(update_batch) == batch_size:
            res_shop = shop_connector.update_shop_product_batch(
                    update_batch, **kwargs)
            # print(res_shop)
            #print(f'last update {product_id} ')
            update_batch = []

    if len(update_batch) > 0:
        res_shop = shop_connector.update_shop_product_batch(
                update_batch, **kwargs)
        # print(res_shop)
        print(f'last update {product_id} ')



def update_product_at_front_shop(table_objs, session, shop_connector, 
                                   data, sku, **kwargs):
    if shop_connector.front_shop_id is None:
        get_front_shop_id(table_objs, session, shop_connector)
    product_id = None
    parent_id = None
    product_info = {}
    if data.get('destination') is None or \
    data['destination'].get('destination_product_id') is None:
        exists = search_exist(table_objs.get('destination'), 
                              session, 
                              {'sku': sku,
                               'front_shop_id': shop_connector.front_shop_id,
                            #    'available': True,
                               'is_current_price': True})
        if exists is None or len(exists) == 0:
            logger.info(f"variation sku {sku} not at shop or not available")
            return None
        product_id = exists[0].destination_product_id
        parent_id = exists[0].destination_parent_id
        if data['destination'].get('price') is None:
            data['destination']['price'] = exists[0].price
    else:
        product_id = data['destination']['destination_product_id']
        parent_id = data['destination'].get('destination_parent_id')
    
    product_info.update(data.get('product', {}))
    product_info.update(data.get('variation', {}))
    product_info.update(data.get('destination', {}))
    #update frontshop 
    res = shop_connector.update_shop_product(
                                   product_info, 
                                   str(product_id), 
                                   parent_id=parent_id,
                                   **kwargs)
    logger.debug(f'{product_id}, {sku}')
    return res

def _get_category_by_name(table_obj, session, name):
    exists = search_exist(table_obj, session, {'category': name})
    if exists and len(exists) > 0:
        return {'category_id': exists[0].id}
    new_cat = add_json_to_db(table_obj, session, {'category': name})
    return {'category_id': new_cat['id']}

# def _update_product_at_frontshop_by_sku(table_objs, session, shop_connector, 
#                                           sku:str, data:dict):
#     """search product at frontshop by sku, 
#        if exist, update
#        Unavailable with shopify
#     """
#     product_id = shop_connector.get_destination_product_id(sku)
#     if product_id is not None:
#         return shop_connector.update_shop_product(
#                                    data, 
#                                    str(product_id),
#                                    formated=True)


from .base_object import BaseObject


class Customer(BaseObject):
    __table_name__ = 'customer'
    def __init__(self, attrs: dict = None):
        super().__init__(attrs)
    
    def set_customer(self, table_objs, attrs: dict):
        self.set_attrs_by_table_class(table_objs.get(self.__table_name__), 
                                     attrs)

    def get_issues(self):
        pass 

    def add_issues(self):
        pass
            
    



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypipet',
 'pypipet.cli',
 'pypipet.cli.cli',
 'pypipet.core',
 'pypipet.core.fileIO',
 'pypipet.core.logging',
 'pypipet.core.model',
 'pypipet.core.operations',
 'pypipet.core.pipeline',
 'pypipet.core.shop_conn',
 'pypipet.core.sql',
 'pypipet.core.transform',
 'pypipet.plugins',
 'pypipet.plugins.canadapost',
 'pypipet.plugins.gg_merchant.shopping',
 'pypipet.plugins.gg_merchant.shopping.content',
 'pypipet.plugins.paypal',
 'pypipet.plugins.woocommerce']

package_data = \
{'': ['*'],
 'pypipet': ['pypipet.egg-info/*'],
 'pypipet.core': ['default_setting/*']}

install_requires = \
['Flask>=2.0.2,<3.0.0',
 'Jinja2>=3.0.3,<4.0.0',
 'ShopifyAPI>=10.0.0,<11.0.0',
 'click>=8.0,<9.0',
 'pandas>=1.2.0,<2.0.0',
 'pyactiveresource>=2.2.2,<3.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.23.0,<3.0.0',
 'sqlalchemy>=1.4.27,<2.0.0']

entry_points = \
{'console_scripts': ['pypipet = pypipet.cli:cli']}

setup_kwargs = {
    'name': 'pypipet',
    'version': '1.1.1',
    'description': 'pypipet',
    'long_description': '\n\n[![GitHub license](https://img.shields.io/github/license/Naereen/StrapDown.js.svg)](https://github.com/Naereen/StrapDown.js/blob/master/LICENSE)\n[![Generic badge](https://img.shields.io/badge/Status-stable-blue.svg)](https://shields.io/)\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://GitHub.com/Naereen/StrapDown.js/graphs/commit-activity)\n[![Generic badge](https://img.shields.io/badge/Pypi-1.0.0-blue.svg)](https://shields.io/)\n[![Generic badge](https://img.shields.io/badge/Python-3.8-blue.svg)](https://shields.io/)\n\n\n# Introduction\n\n**`PyPipet`** is an open source project, aiming to integrate data flows of Ecommerce. It provides platform-independent data flows to support Ecommerce functionality. It simplifies data pipelines of data management in ecommerce, for example, adding catalog, updating product, managing inventory and orders, etc. It is specially customized for small business who are selling on wordpress (for example, with woocommerce), shopify, ebay, etc. (more frontshop connected will be added). It extremely handy if the business is selling on multiple platforms (e.g., for dropshipping). It makes it extreamly easy to move your frontshop from one platform to anothor.\n\n* For source code,  visit  [github repository](https://github.com/pypipet/pypipet).\n* For documentation, vist [docs](https://pypipet.com)\n\n## Latest `pip` [version: 1.1.1](https://pypi.org/project/pypipet/)\n\n### [Change Logs](https://pypi.org/project/changelogs)\n\n[reporting bugs here](https://github.com/pypipet/pypipet/issues)\n\n## [Dependencies](https://pypipet.com/dependencies)\n\n# Installation\n\nuse `pip`\n\n    pip install -r requirements.txt\n\n    pip install --upgrade pypipet  \n\n\nfor using Google Content API to connect Google Merchant, please install\n[`google_auth_httplib2`](https://pypi.org/project/google-auth-httplib2/)\nand [`googleapiclient`](https://github.com/googleapis/google-api-python-client)\n\n## Tested with\n\n- Shopify\n- Woocommerce with Wordpress.com\n- Postgresql\n- AWS RDS\n\n\n## Key features\n\n- cli supported\n- catalog import/export\n- product management for publishing\n- order management\n- fulfillment management\n- inventory management with multiple suppliers\n\n# [Quick start guide](https://pypipet.com/quick_start/create_project/)\n\n\n# [Use cases](https://pypipet.com/usecases/usecases)\n\n\n\n\n\n\n',
    'author': 'pypipet and contributors',
    'author_email': 'pypipet@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import click 
from .cli import cli
import pandas as pd

@cli.command()
@click.option("--file", type=click.Choice(['product', 'product_update',
                     'order', 'customer', 'inventory', 'fulfillment',
                    'supplier', 'front_shop', 'tax', 'category' ]))
@click.pass_context
def template(ctx,file):
    """
    generate csv template for importing
    """
    click.echo('generate template: {}'.format(file))
    project = ctx.obj['project']
    config = ctx.obj["config"]
    if config is None:
        click.secho('create project first',
                   fg='yellow')
        return

    project.initialize_project(config)
    file_template = project.load_file_template()
    template = None
    if file == 'product':
        template = file_template.get('product_info') 
    elif file == 'product_update':
        template = file_template.get('product_update') 
    elif file == 'order':
        pass 
    elif file == 'customer':
        pass 
    elif file == 'inventory':
        template = file_template.get('inventory')  
    elif file == 'fulfillment':
        template = file_template.get('fulfillment')  
    elif file == 'supplier':
        template = file_template.get('supplier')
    elif file == 'front_shop':
        template = file_template.get('front_shop')
    elif file == 'tax':
        template = file_template.get('tax')
    elif file == 'category':
        template = file_template.get('category')
    else:
        click.echo('file template: {} not supported'.format(file))
        return 
    
    cols = []
    if type(template) is dict:
        for table, fields in template.items():
            click.echo(fields)
            fields = [table+'.'+ f for f in fields]
            cols += list(fields)
    else:
        click.echo(template)
        cols = template

    df=pd.DataFrame(columns=cols)
    df.to_csv(project.root + file + '_template.csv', index=False)
    


import click 
from .cli import cli
from pypipet.core.operations.utility import get_front_shop_id
from pypipet.core.operations.inventory import update_inventory_bulk
from pypipet.core.operations.inventory import update_instock_qty_db
from pypipet.core.operations.inventory import update_instock_front_shop
# from pypipet.core.operations.inventory import update_instock_qty_db_simple
from pypipet.core.operations.inventory import get_inventory_by_sku
from pypipet.core.operations.inventory import update_instock_front_shop_by_sku
from pypipet.core.fileIO.file_loader import read_csv_file
from .utility import col2dict_update
from pprint import pprint


#do do: add new inventory, view inentory

@cli.command()
@click.argument("action", type=click.Choice(['show', 'edit']))
@click.option("--shop", 
                help="frontshop name (in setting.shops)")
@click.option("--sku",
                help="if multiple sku provided, seperate with comma")
@click.option("--qty", type=int, 
                help="total qty by sku")
@click.option("-f", "--filename",
                help="update inventory with file")
@click.option("--batch", default=100,
                help="batch size to update inventory in database")
@click.option('--skip-shop', is_flag=True, 
                  help="skip updating front shop for action edit.")
@click.option('--ignore-new', is_flag=True, 
                  help="ignore new skus")
@click.pass_context
def inventory(ctx, action, shop, filename, batch, 
                sku,qty, skip_shop, ignore_new):
    """
    add/edit inventory (by skus or by file)
    """
    click.echo(' inventory at... {}'.format(shop))

    project = ctx.obj['project']
    config = ctx.obj["config"]
    if config is None:
        click.secho('create project first',
                   fg='yellow')
        return

    project.initialize_project(config)
    table_classes = project.get_table_objects()
    sessionmaker = project.get_session_maker()
    session = sessionmaker()
    shop_conn = project.get_shop_connector(shop)
    if shop_conn is None:
        click.secho('shop setting not found: {}'.format(shop),
                   fg='yellow')
        session.close()
        return 
    
    get_front_shop_id(table_classes, session, shop_conn)
    
    if action == 'edit':
        click.echo('updating inventory, it may take a while \
        (depends on the feed size)')
        _edit_inventory(session, table_classes, shop_conn, filename, 
                        batch,sku,qty, skip_shop, ignore_new)
    elif action == 'show':
        if sku is None:
            click.secho('missing sku',
                fg='yellow')
        sku = sku.split(',')
        for s in sku:
            inv = get_inventory_by_sku(table_classes, session, s, 
                                            by_supplier=True)
            click.echo(f"sku {s} in_stock_qty: {inv}")
    
    session.close()


def _edit_inventory(session, table_classes, shop_conn, filename, 
              batch, sku,qty, skip_shop, ignore_new):
    if filename is not None:
        invs = _get_data_from_file(filename) 
        update_inventory_bulk(table_classes, 
                             session, invs, ignore_new=ignore_new)
        
        #if not skip, update shop
        if not skip_shop:
            update_instock_front_shop(table_classes, session, shop_conn,
                                batch_size=batch,latest_hours=1)
    else:
        click.secho('missing filename', fg='yellow')



def _get_data_from_file(filename):
    data = read_csv_file(filename)
    invs = []
    #validation
    for i, row in data.iterrows():
        if row['sku'].strip() == '':
            click.secho('missing sku in row {}, stop processing'\
                                 .format(i+2), 
                        fg='yellow')
            return 
    
    for i, row in data.iterrows():
        inv = {}
        for k, val in dict(row).items():
            if val.strip() == '':
                continue 
            if k in ['qty', 'supplier_id']: 
                val = int(val)
            if 'cost' in k or 'price' in k: 
                val = float(val) 
            inv[k] = val
        # pprint(inv)
        invs.append(inv)

    return invs


from .base_object import BaseObject
from passlib.hash import pbkdf2_sha256 as sha256


class User(BaseObject):
    __table_name__ = 'login_user'
    def __init__(self, attrs: dict = None):
        super().__init__(attrs)

    def set_user(self, table_objs, attrs: dict):
        self.set_attrs_by_table_class(table_objs.get(self.__table_name__), 
                                     attrs)

    # @classmethod
    # def find_by_username(cls, username):
    #     return 

    @staticmethod
    def generate_hash(password):
        return sha256.hash(password)

    @staticmethod
    def verify_hash(password, hash):
        return sha256.verify(password, hash)
